import json
import shutil
from pathlib import Path

from .naming import normalize_name


def scan_dist_info(site_pkg):
    issues = []
    dist_infos = []
    for p in Path(site_pkg).iterdir():
        if p.is_dir() and p.name.endswith(".dist-info"):
            dist_infos.append(p.name)

    versions = {}
    paths = {}
    for d in dist_infos:
        base = d[:-len(".dist-info")]
        if "-" not in base:
            continue
        name, version = base.rsplit("-", 1)
        key = normalize_name(name)
        versions.setdefault(key, set()).add(version)
        paths.setdefault(key, []).append(str(Path(site_pkg) / d))

    for pkg, vers in sorted(versions.items()):
        if len(vers) > 1:
            issues.append(
                {
                    "type": "duplicate-dist-info",
                    "package": pkg,
                    "versions": sorted(vers),
                    "paths": paths.get(pkg, []),
                }
            )
    return issues


def scan_pyd_duplicates(site_pkg):
    issues = []
    pyd_files = [p.name for p in Path(site_pkg).glob("*.pyd")]
    base_map = {}
    for name in pyd_files:
        stem = name[:-4]
        base = stem.split(".cp")[0]
        base_map.setdefault(base, []).append(name)
    for base, files in sorted(base_map.items()):
        if len(files) > 1:
            issues.append({"type": "duplicate-pyd", "base": base, "files": sorted(files)})
    return issues


def scan_invalid_artifacts(site_pkg):
    issues = []
    root = Path(site_pkg)
    if not root.exists():
        return issues
    for p in root.iterdir():
        name = p.name
        if name.startswith("~") or name.endswith(".conda_trash"):
            issues.append({"type": "invalid-artifact", "path": str(p), "name": name})
    return issues


def remove_invalid_artifact(path):
    p = Path(path)
    if not p.exists():
        return True
    try:
        if p.is_dir():
            shutil.rmtree(p)
        else:
            p.unlink()
        return True
    except OSError:
        return False


def remove_dist_info_paths(paths):
    ok = True
    for path in paths:
        p = Path(path)
        if not p.exists():
            continue
        try:
            if p.is_dir():
                shutil.rmtree(p)
            else:
                p.unlink()
        except OSError:
            ok = False
    return ok


def parse_conda_meta_filename(filename):
    """
    Parse conda-meta json filename like: <name>-<version>-<build>.json
    Returns (name, version, build) or (None, None, None) if not parseable.
    """
    name = Path(filename).name
    if not name.endswith(".json"):
        return None, None, None
    base = name[:-5]
    parts = base.rsplit("-", 2)
    if len(parts) != 3:
        return None, None, None
    pkg_name, version, build = parts
    if not pkg_name or not version or not build:
        return None, None, None
    return pkg_name, version, build


def scan_conda_meta_json(env_path):
    """
    Scan <env>/conda-meta/*.json for broken records.
    Current checks:
      - invalid JSON
      - missing required key: depends
    """
    issues = []
    root = Path(env_path) / "conda-meta"
    if not root.exists():
        return issues
    for p in root.glob("*.json"):
        pkg_name, version, build = parse_conda_meta_filename(p.name)
        try:
            data = json.loads(p.read_text(encoding="utf-8", errors="strict"))
        except Exception:
            issues.append(
                {
                    "type": "conda-meta-invalid-json",
                    "path": str(p),
                    "package": pkg_name,
                    "version": version,
                    "build": build,
                }
            )
            continue
        if not isinstance(data, dict):
            issues.append(
                {
                    "type": "conda-meta-invalid-json",
                    "path": str(p),
                    "package": pkg_name,
                    "version": version,
                    "build": build,
                }
            )
            continue
        if "depends" not in data:
            # Some conda-meta records legitimately omit `depends` (e.g. certain noarch/helper packages).
            # Only flag as broken if the record also looks incomplete.
            required = ("name", "version", "build")
            if not all(k in data for k in required):
                issues.append(
                    {
                        "type": "conda-meta-missing-depends",
                        "path": str(p),
                        "package": pkg_name,
                        "version": version,
                        "build": build,
                    }
                )
    return issues



import unittest

from env_repair.inconsistent import parse_inconsistent


class TestInconsistentParse(unittest.TestCase):
    def test_detects_inconsistent(self):
        txt = "Warning: The environment is inconsistent, please check.\n- numpy\n- scipy\n"
        ok, pkgs = parse_inconsistent(txt)
        self.assertTrue(ok)
        self.assertEqual(set(pkgs), {"numpy", "scipy"})

    def test_no_inconsistent(self):
        ok, pkgs = parse_inconsistent("All good")
        self.assertFalse(ok)
        self.assertEqual(pkgs, [])


if __name__ == "__main__":
    unittest.main()




import argparse
import json
import os
import sys

from .doctor import (
    cache_check,
    cache_fix,
    diagnose_clobber,
    diagnose_inconsistent,
    diagnose_ssl,
    fix_inconsistent,
    rebuild,
    rollback,
    run,
)
from .verify_imports import verify_imports
from .subprocess_utils import OperationInterrupted
from .i18n import t


def build_parser():
    lang = "auto"
    p = argparse.ArgumentParser(
        prog="env-repair",
        description=t("help_desc", lang=lang),
        add_help=False,
    )
    p.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    sub = p.add_subparsers(dest="cmd")

    rb = sub.add_parser(
        "rollback",
        help=t("help_cmd_rollback", lang=lang),
        description=t("help_cmd_rollback", lang=lang),
        add_help=False,
    )
    rb.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    rb.add_argument("--env", required=True, help=t("help_env_single", lang=lang))
    rb.add_argument("--to", default="prev", help=t("help_to", lang=lang))
    rb.add_argument("--dry-run", action="store_true", help=t("help_dry_run", lang=lang))
    rb.add_argument("-y", "--yes", action="store_true", help=t("help_yes", lang=lang))
    rb.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    rb.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))
    rb.add_argument("--plan", action="store_true", help=t("help_plan", lang=lang))

    rb2 = sub.add_parser(
        "rebuild",
        help=t("help_cmd_rebuild", lang=lang),
        description=t("help_cmd_rebuild", lang=lang),
        add_help=False,
    )
    rb2.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    rb2.add_argument("--env", required=True, help=t("help_env_single", lang=lang))
    rb2.add_argument("--to", required=True, help=t("help_to_rebuild", lang=lang))
    rb2.add_argument("--verify", action="store_true", help=t("help_verify", lang=lang))
    rb2.add_argument("--plan", action="store_true", help=t("help_plan", lang=lang))
    rb2.add_argument("-y", "--yes", action="store_true", help=t("help_yes", lang=lang))
    rb2.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    rb2.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    dc = sub.add_parser(
        "diagnose-clobber",
        help=t("help_cmd_diagnose_clobber", lang=lang),
        description=t("help_cmd_diagnose_clobber", lang=lang),
        add_help=False,
    )
    dc.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    dc.add_argument("--env", required=True, help=t("help_env_single", lang=lang))
    dc.add_argument("--logfile", required=True, help=t("help_logfile", lang=lang))
    dc.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    dc.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    di = sub.add_parser(
        "diagnose-inconsistent",
        help=t("help_cmd_diagnose_inconsistent", lang=lang),
        description=t("help_cmd_diagnose_inconsistent", lang=lang),
        add_help=False,
    )
    di.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    di.add_argument("--env", required=True, help=t("help_env_single", lang=lang))
    di.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    di.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    fi = sub.add_parser(
        "fix-inconsistent",
        help=t("help_cmd_fix_inconsistent", lang=lang),
        description=t("help_cmd_fix_inconsistent", lang=lang),
        add_help=False,
    )
    fi.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    fi.add_argument("--env", required=True, help=t("help_env_single", lang=lang))
    fi.add_argument("--level", choices=["safe", "normal", "rebuild"], default="safe", help=t("help_level_inconsistent", lang=lang))
    fi.add_argument("--plan", action="store_true", help=t("help_plan", lang=lang))
    fi.add_argument("-y", "--yes", action="store_true", help=t("help_yes", lang=lang))
    fi.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    fi.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    cc = sub.add_parser(
        "cache-check",
        help=t("help_cmd_cache_check", lang=lang),
        description=t("help_cmd_cache_check", lang=lang),
        add_help=False,
    )
    cc.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    cc.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    cc.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    cf = sub.add_parser(
        "cache-fix",
        help=t("help_cmd_cache_fix", lang=lang),
        description=t("help_cmd_cache_fix", lang=lang),
        add_help=False,
    )
    cf.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    cf.add_argument("--level", choices=["safe", "targeted", "aggressive"], default="safe", help=t("help_level_cache", lang=lang))
    cf.add_argument("--plan", action="store_true", help=t("help_plan", lang=lang))
    cf.add_argument("-y", "--yes", action="store_true", help=t("help_yes", lang=lang))
    cf.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    cf.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    ds = sub.add_parser(
        "diagnose-ssl",
        help=t("help_cmd_diagnose_ssl", lang=lang),
        description=t("help_cmd_diagnose_ssl", lang=lang),
        add_help=False,
    )
    ds.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    ds.add_argument("--env", help=t("help_env_single", lang=lang))
    ds.add_argument("--base", action="store_true", help=t("help_base", lang=lang))
    ds.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    ds.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))

    vi = sub.add_parser(
        "verify-imports",
        help="Check if installed packages can be imported",
        description="Verify that packages are importable (finds missing DLLs, etc.)",
        add_help=False,
    )
    vi.add_argument("-h", "--help", action="help", help=t("help_help", lang=lang))
    vi.add_argument("--env", help=t("help_env_single", lang=lang))
    vi.add_argument("--full", action="store_true", help="Check all packages (default: critical only)")
    vi.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    vi.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))
    vi.add_argument("--fix", action="store_true", help="Attempt to automatically fix broken imports")

    p.add_argument(
        "--env",
        action="append",
        default=[],
        help=t("help_env_multi", lang=lang),
    )
    p.add_argument("--fix", action="store_true", help=t("help_fix", lang=lang))
    p.add_argument("--adopt-pip", action="store_true", help=t("help_adopt_pip", lang=lang))
    p.add_argument(
        "--keep-pip",
        action="store_true",
        help=t("help_keep_pip", lang=lang),
    )
    p.add_argument(
        "--prefer",
        choices=["auto", "conda", "pip"],
        default="auto",
        help=t("help_prefer", lang=lang),
    )
    p.add_argument("--pip-fallback", action="store_true", help=t("help_pip_fallback", lang=lang))
    p.add_argument("--no-pip-fallback", action="store_true", help=t("help_no_pip_fallback", lang=lang))
    p.add_argument("--channel", action="append", default=[], help=t("help_channel", lang=lang))
    p.add_argument("--no-channels-from-condarc", action="store_true", help=t("help_no_channels_from_condarc", lang=lang))
    p.add_argument("--no-default-channels", action="store_true", help=t("help_no_default_channels", lang=lang))
    p.add_argument("--ignore-pinned", action="store_true", help=t("help_ignore_pinned", lang=lang))
    p.add_argument("--force-reinstall", action="store_true", help=t("help_force_reinstall", lang=lang))
    p.add_argument("--snapshot", help=t("help_snapshot", lang=lang))
    p.add_argument("--json", action="store_true", help=t("help_json", lang=lang))
    p.add_argument("--debug", action="store_true", help=t("help_debug", lang=lang))
    return p


def main(argv=None):
    parser = build_parser()
    args = parser.parse_args(argv)
    if getattr(args, "debug", False):
        # Make subprocess executions print conda/mamba command lines for transparency.
        os.environ["ENV_REPAIR_SHOW_CMDS"] = "1"
    try:
        if args.cmd == "rollback":
            result = rollback(args)
        elif args.cmd == "rebuild":
            result = rebuild(args)
        elif args.cmd == "diagnose-clobber":
            result = diagnose_clobber(args)
        elif args.cmd == "diagnose-inconsistent":
            result = diagnose_inconsistent(args)
        elif args.cmd == "fix-inconsistent":
            result = fix_inconsistent(args)
        elif args.cmd == "cache-check":
            result = cache_check(args)
        elif args.cmd == "cache-fix":
            result = cache_fix(args)
        elif args.cmd == "diagnose-ssl":
            result = diagnose_ssl(args)
        elif args.cmd == "verify-imports":
            result = verify_imports(args)
        else:
            result = run(args)
    except (KeyboardInterrupt, OperationInterrupted):
        print(t("interrupted", lang="auto"), file=sys.stderr)
        return 130
    if getattr(args, "json", False):
        print(json.dumps(result.get("report"), indent=2))
    return int(result.get("exit_code", 0 if result.get("ok") else 1))


if __name__ == "__main__":
    raise SystemExit(main())



import unittest

from env_repair.conda_ops import extract_pkgs_dirs


class TestExtractPkgsDirs(unittest.TestCase):
    def test_extracts_list(self):
        info = {"pkgs_dirs": ["C:\\Anaconda3\\pkgs", "D:\\cache\\pkgs"]}
        self.assertEqual(extract_pkgs_dirs(info), ["C:\\Anaconda3\\pkgs", "D:\\cache\\pkgs"])

    def test_handles_missing(self):
        self.assertEqual(extract_pkgs_dirs({}), [])


if __name__ == "__main__":
    unittest.main()




import unittest
from unittest import mock
import io
import contextlib

from env_repair.subprocess_utils import OperationInterrupted


class TestCliInterrupt(unittest.TestCase):
    def test_cli_catches_operation_interrupted(self):
        import env_repair.cli as cli

        with (
            mock.patch.object(cli, "run", side_effect=OperationInterrupted(["mamba", "search"], returncode=130)),
            contextlib.redirect_stderr(io.StringIO()),
        ):
            rc = cli.main(["--env", "base"])
        self.assertEqual(rc, 130)


if __name__ == "__main__":
    unittest.main()



import re


def normalize_name(name):
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_name_simple(name):
    return re.sub(r"[^a-z0-9]+", "", name.lower())


def build_search_variants(name):
    variants = {name}
    if "-" in name:
        variants.add(name.replace("-", "."))
        variants.add(name.replace("-", "_"))
        variants.add(name.replace("-", ""))
    if "_" in name:
        variants.add(name.replace("_", "-"))
        variants.add(name.replace("_", "."))
        variants.add(name.replace("_", ""))
    if "." in name:
        variants.add(name.replace(".", "-"))
        variants.add(name.replace(".", "_"))
        variants.add(name.replace(".", ""))
    return [v for v in variants if v]




import locale
import os


MESSAGES = {
    "en": {
        "interrupted": "Interrupted.",
        "prompt_interrupted": "Interrupted. (r)estore from snapshot, (c)ontinue, (a)bort? [a]: ",
        "prompt_approve": "Proceed? [y/N]: ",
        "abort": "Aborted.",
        "manager_missing": "No conda/mamba/micromamba found in PATH; some actions are unavailable.",
        "rollback_plan": "rollback plan: {from_rev} -> {to_rev}",
        "rollback_no_revisions": "no conda revisions found for this environment",
        "rollback_invalid_target": "invalid rollback target: {to}",
        "rollback_done": "rollback done: {to_rev}",
        "rebuild_plan": "rebuild plan: {src} -> {dst}",
        "rebuild_done": "rebuild done: {dst}",
        "rebuild_target_exists": "rebuild target already exists: {dst}",
        "rebuild_need_yes": "rebuild requires confirmation (use -y/--yes to skip prompt)",
        "plan_only": "plan-only (no changes applied)",
        "clobber_header": "clobber diagnosis:",
        "clobber_no_log": "no logfile provided",
        "clobber_no_paths": "no conflict paths found in log",
        "clobber_log_read_failed": "failed to read logfile: {path}",
        "inconsistent_header": "inconsistent diagnosis:",
        "inconsistent_found": "environment is inconsistent",
        "inconsistent_not_found": "no inconsistency warning detected",
        "inconsistent_fix_plan": "fix-inconsistent plan (level={level}): {env}",
        "inconsistent_fix_done": "fix-inconsistent done (level={level})",
        "cache_header": "cache check:",
        "cache_pkgs_dirs": "pkgs_dirs: {value}",
        "cache_fix_plan": "cache-fix plan (level={level})",
        "cache_fix_done": "cache-fix done (level={level})",
        "ssl_header": "ssl diagnosis:",
        "ssl_ok": "ssl import ok: {value}",
        "ssl_fail": "ssl import failed: {value}",
        "ssl_hint": "hint: try `mamba install -n base openssl ca-certificates certifi` (or `conda install ...`), and ensure proper conda activation in your shell",
        "help_desc": "Scan and repair conda/mamba/micromamba environments with mixed conda/pip installs.",
        "help_help": "Show this help message and exit.",
        "help_env_multi": "Environment name(s) or path(s). Repeatable. Default: all discovered envs.",
        "help_fix": "Attempt repairs (safe-ish).",
        "help_adopt_pip": "Try to replace pip-installed packages with conda ones.",
        "help_keep_pip": "Keep the original pip-installed packages even if adopt-pip succeeds (default: uninstall pip version).",
        "help_prefer": "Preferred installer when fixing duplicates.",
        "help_pip_fallback": "Allow pip reinstall if conda reinstall fails.",
        "help_no_pip_fallback": "Disable pip fallback.",
        "help_channel": "Add conda channel(s). Repeatable.",
        "help_no_channels_from_condarc": "Ignore channels from .condarc/conda config.",
        "help_no_default_channels": "Do not auto-add defaults/anaconda.",
        "help_ignore_pinned": "Ignore conda pinned specs during installs.",
        "help_force_reinstall": "Force reinstall packages via conda.",
        "help_snapshot": "Write env export YAML to this path (conda envs only).",
        "help_json": "Output machine-readable JSON.",
        "help_debug": "Verbose debug logging + show JSON tool output.",
        "help_cmd_rollback": "Rollback a conda env to an earlier revision.",
        "help_cmd_rebuild": "Rebuild an env into a new env (export/import).",
        "help_cmd_diagnose_clobber": "Parse a conda/mamba ClobberError log and find file owners.",
        "help_cmd_diagnose_inconsistent": 'Check whether an env is "inconsistent" (conda warning).',
        "help_cmd_fix_inconsistent": 'Attempt to repair an "inconsistent" env.',
        "help_cmd_cache_check": "Show conda package cache locations.",
        "help_cmd_cache_fix": "Clean conda package caches.",
        "help_cmd_diagnose_ssl": "Diagnose SSL/OpenSSL issues (advisor).",
        "help_env_single": "Environment name or path.",
        "help_to": "Target: prev|latest|<N> (default: prev).",
        "help_dry_run": "Only simulate the action.",
        "help_yes": "Do not ask for confirmation.",
        "help_plan": "Only show plan (no changes).",
        "help_to_rebuild": "Target env name or path.",
        "help_verify": "Scan the new env after creation.",
        "help_logfile": "Path to conda/mamba output log text.",
        "help_level_inconsistent": "Fix level.",
        "help_level_cache": "Clean level.",
        "help_base": "Diagnose base/root prefix.",
    "progress_envs": "envs",
        "step_scan": "scan",
        "step_snapshot": "snapshot",
        "step_fix": "fix",
        "step_adopt_pip": "adopt-pip",
        "adopt_search": "mamba search batch",
        "adopt_resolve": "adopt-pip resolve",
        "env_header": "== {name} ==",
        "path": "path: {value}",
        "python": "python: {value}",
        "channels": "channels: {value}",
        "snapshot": "snapshot: {path} {status}",
        "snapshot_ok": "ok",
        "snapshot_failed": "failed",
        "issues_none": "issues: none",
        "issues": "issues:",
        "pinned": "pinned:",
        "fixes_none": "fixes: none",
        "fixes": "fixes:",
        "issue_duplicate_dist_info": " - duplicate-dist-info: {package} {versions}",
        "issue_duplicate_pyd": " - duplicate-pyd: {base} {files}",
        "issue_invalid_artifact": " - invalid-artifact: {name}",
        "issue_generic": " - {type}",
        "fix_line": " - {label} {method} {status}",
        "fix_ok": "ok",
        "fix_failed": "failed",
        "fix_report": "fix report:",
        "fix_report_none": "no fixes applied",
        "col_action": "action",
        "col_item": "item",
        "col_result": "result",
        "col_reason": "reason",
        "reason_stale_artifact": "stale/invalid artifact in site-packages",
        "reason_duplicate_dist_info": "duplicate .dist-info entries detected",
        "reason_conda_meta_reinstall": "broken conda-meta record(s) (missing depends / invalid json)",
        "reason_case_conflict_pip_uninstall": "pip+conda installed same version with different name casing",
        "reason_case_conflict_conda_relink": "relink conda package after removing pip duplicate",
        "reason_reinstall_duplicates": "reinstall to resolve duplicates",
        "reason_adopt_conda_install": "install mapped conda packages for pip-installed candidates",
        "reason_adopt_skip_keep": "keep {pip} (pip {pip_version}) because conda {conda} version is {conda_version}",
        "reason_adopt_pip_uninstall": "remove pip duplicates after conda install (only when versions match)",
        "reason_adopt_pip_force_uninstall": "remove pip package because a conda replacement is present",
        "reason_adopt_conda_relink": "relink conda packages after pip uninstall",
    },
    "de": {
        "interrupted": "Abgebrochen.",
        "prompt_interrupted": "Abgebrochen. (r)estore aus Snapshot, (c)ontinue, (a)bort? [a]: ",
        "prompt_approve": "Fortfahren? [j/N]: ",
        "abort": "Abgebrochen.",
        "manager_missing": "Kein conda/mamba/micromamba im PATH gefunden; manche Aktionen sind nicht verfuegbar.",
        "rollback_plan": "Rollback-Plan: {from_rev} -> {to_rev}",
        "rollback_no_revisions": "keine conda Revisions fuer dieses Environment gefunden",
        "rollback_invalid_target": "ungueltiges Rollback-Ziel: {to}",
        "rollback_done": "Rollback fertig: {to_rev}",
        "rebuild_plan": "Rebuild-Plan: {src} -> {dst}",
        "rebuild_done": "Rebuild fertig: {dst}",
        "rebuild_target_exists": "Rebuild-Ziel existiert bereits: {dst}",
        "rebuild_need_yes": "Rebuild braucht Bestaetigung (mit -y/--yes ohne Rueckfrage)",
        "plan_only": "nur Plan (keine Aenderungen)",
        "clobber_header": "Clobber-Diagnose:",
        "clobber_no_log": "keine Logdatei angegeben",
        "clobber_no_paths": "keine Konfliktpfade im Log gefunden",
        "clobber_log_read_failed": "Logdatei konnte nicht gelesen werden: {path}",
        "inconsistent_header": "Inconsistent-Diagnose:",
        "inconsistent_found": "Environment ist inconsistent",
        "inconsistent_not_found": "keine Inconsistency-Warnung erkannt",
        "inconsistent_fix_plan": "Fix-Inconsistent Plan (level={level}): {env}",
        "inconsistent_fix_done": "Fix-Inconsistent fertig (level={level})",
        "cache_header": "Cache-Check:",
        "cache_pkgs_dirs": "pkgs_dirs: {value}",
        "cache_fix_plan": "Cache-Fix Plan (level={level})",
        "cache_fix_done": "Cache-Fix fertig (level={level})",
        "ssl_header": "SSL-Diagnose:",
        "ssl_ok": "ssl import ok: {value}",
        "ssl_fail": "ssl import fehlgeschlagen: {value}",
        "ssl_hint": "Hinweis: ggf. `mamba install -n base openssl ca-certificates certifi` (oder `conda install ...`) und korrekte conda Aktivierung im Terminal pruefen",
        "help_desc": "Scan und Repair fuer conda/mamba/micromamba Environments mit gemischten conda/pip Installationen.",
        "help_help": "Hilfe anzeigen und beenden.",
        "help_env_multi": "Environment Name(n) oder Pfad(e). Wiederholbar. Default: alle gefundenen Envs.",
        "help_fix": "Reparaturen ausfuehren (safe-ish).",
        "help_adopt_pip": "Versuche pip-installierte Pakete durch conda Pakete zu ersetzen.",
        "help_keep_pip": "pip-Version behalten, auch wenn adopt-pip erfolgreich ist (Default: pip-Version entfernen).",
        "help_prefer": "Bevorzugter Installer beim Fixen von Duplikaten.",
        "help_pip_fallback": "pip Reinstall erlauben, wenn conda/mamba fehlschlaegt.",
        "help_no_pip_fallback": "pip Fallback deaktivieren.",
        "help_channel": "Conda Channel(s) hinzufuegen. Wiederholbar.",
        "help_no_channels_from_condarc": "Channels aus .condarc/conda config ignorieren.",
        "help_no_default_channels": "defaults/anaconda nicht automatisch hinzufuegen.",
        "help_ignore_pinned": "Conda Pins (pinned specs) beim Install ignorieren.",
        "help_force_reinstall": "Pakete via conda/mamba erneut installieren (force-reinstall).",
        "help_snapshot": "Env-Export YAML nach diesem Pfad schreiben (nur conda Envs).",
        "help_json": "Maschinenlesbares JSON ausgeben.",
        "help_debug": "Debug-Ausgabe + JSON Tool-Output anzeigen.",
        "help_cmd_rollback": "Rollback eines conda Envs auf eine fruehere Revision.",
        "help_cmd_rebuild": "Rebuild: Env exportieren und als neues Env neu erstellen.",
        "help_cmd_diagnose_clobber": "ClobberError Log parsen und Datei-Owner bestimmen.",
        "help_cmd_diagnose_inconsistent": 'Pruefen, ob ein Env "inconsistent" ist (conda Warnung).',
        "help_cmd_fix_inconsistent": 'Versuch, ein "inconsistent" Env zu reparieren.',
        "help_cmd_cache_check": "Conda Package Cache Locations anzeigen.",
        "help_cmd_cache_fix": "Conda Package Caches bereinigen.",
        "help_cmd_diagnose_ssl": "SSL/OpenSSL Probleme diagnostizieren (Advisor).",
        "help_env_single": "Environment Name oder Pfad.",
        "help_to": "Ziel: prev|latest|<N> (Default: prev).",
        "help_dry_run": "Aktion nur simulieren.",
        "help_yes": "Keine Rueckfrage (auto bestaetigen).",
        "help_plan": "Nur Plan anzeigen (keine Aenderungen).",
        "help_to_rebuild": "Ziel-Environment Name oder Pfad.",
        "help_verify": "Neues Env nach Erstellung scannen.",
        "help_logfile": "Pfad zur conda/mamba Output-Logdatei.",
        "help_level_inconsistent": "Fix-Level.",
        "help_level_cache": "Clean-Level.",
        "help_base": "Base/Root Prefix diagnostizieren.",
        "progress_envs": "Envs",
        "step_scan": "Scan",
        "step_snapshot": "Snapshot",
        "step_fix": "Fix",
        "step_adopt_pip": "Adopt-Pip",
        "adopt_search": "mamba search batch",
        "adopt_resolve": "adopt-pip resolve",
        "env_header": "== {name} ==",
        "path": "Pfad: {value}",
        "python": "Python: {value}",
        "channels": "Channels: {value}",
        "snapshot": "Snapshot: {path} {status}",
        "snapshot_ok": "ok",
        "snapshot_failed": "fehlgeschlagen",
        "issues_none": "Probleme: keine",
        "issues": "Probleme:",
        "pinned": "Pinned:",
        "fixes_none": "Fixes: keine",
        "fixes": "Fixes:",
        "issue_duplicate_dist_info": " - duplicate-dist-info: {package} {versions}",
        "issue_duplicate_pyd": " - duplicate-pyd: {base} {files}",
        "issue_invalid_artifact": " - invalid-artifact: {name}",
        "issue_generic": " - {type}",
        "fix_line": " - {label} {method} {status}",
        "fix_ok": "ok",
        "fix_failed": "fehlgeschlagen",
        "fix_report": "Fix-Report:",
        "fix_report_none": "keine Fixes ausgefuehrt",
        "col_action": "aktion",
        "col_item": "element",
        "col_result": "ergebnis",
        "col_reason": "grund",
        "reason_stale_artifact": "veraltetes/ungueltiges Artefakt in site-packages",
        "reason_duplicate_dist_info": "doppelte .dist-info Eintraege erkannt",
        "reason_conda_meta_reinstall": "defekte conda-meta Records (depends fehlt / JSON ungueltig)",
        "reason_case_conflict_pip_uninstall": "pip+conda haben gleiche Version mit unterschiedlicher Schreibweise installiert",
        "reason_case_conflict_conda_relink": "conda Paket nach pip-Uninstall erneut relinken",
        "reason_reinstall_duplicates": "Reinstall um Duplikate zu beheben",
        "reason_adopt_conda_install": "gemappte conda Pakete fuer pip-Kandidaten installieren",
        "reason_adopt_skip_keep": "{pip} behalten (pip {pip_version}), weil conda {conda} Version {conda_version} hat",
        "reason_adopt_pip_uninstall": "pip-Duplikate nach conda Install entfernen (nur wenn Versionen matchen)",
        "reason_adopt_pip_force_uninstall": "pip-Paket entfernen, weil conda-Ersatz bereits vorhanden ist",
        "reason_adopt_conda_relink": "conda Pakete nach pip-Uninstall erneut relinken",
    },
}


def detect_lang():
    # Best-effort auto-detection.
    for var in ("LC_ALL", "LC_MESSAGES", "LANG"):
        val = os.environ.get(var)
        if isinstance(val, str) and val:
            if val.lower().startswith("de"):
                return "de"
            return "en"
    try:
        loc = locale.getdefaultlocale()[0]  # noqa: UP010 (keep compat)
    except Exception:
        loc = None
    if isinstance(loc, str) and loc.lower().startswith("de"):
        return "de"
    return "en"


def get_lang(lang):
    if not lang or lang == "auto":
        return detect_lang()
    if lang in MESSAGES:
        return lang
    return "en"


def t(key, *, lang, **kwargs):
    lang = get_lang(lang)
    msg = MESSAGES.get(lang, {}).get(key) or MESSAGES["en"].get(key) or key
    try:
        return msg.format(**kwargs)
    except Exception:
        return msg



import unittest

from env_repair.repair import _adopt_pip_force_uninstall_plan


class TestAdoptPipForceUninstall(unittest.TestCase):
    def test_force_uninstall_pysha3_when_safe_present(self):
        pip_versions = {"pysha3": "1.0.2"}
        entries = [
            {"name": "safe-pysha3", "version": "1.0.4", "channel": "conda-forge"},
            {"name": "pysha3", "version": "1.0.2", "channel": "pypi"},
        ]
        pip_rm, relink, details = _adopt_pip_force_uninstall_plan(pip_versions=pip_versions, entries=entries)
        self.assertEqual(pip_rm, ["pysha3"])
        self.assertEqual(relink, ["safe-pysha3"])
        self.assertEqual(len(details), 1)

    def test_no_force_uninstall_if_replacement_missing(self):
        pip_versions = {"pysha3": "1.0.2"}
        entries = [{"name": "pysha3", "version": "1.0.2", "channel": "pypi"}]
        pip_rm, relink, details = _adopt_pip_force_uninstall_plan(pip_versions=pip_versions, entries=entries)
        self.assertEqual(pip_rm, [])
        self.assertEqual(relink, [])
        self.assertEqual(details, [])


if __name__ == "__main__":
    unittest.main()




import unittest

from env_repair.repair import (
    PYPI_TO_CONDA_NAME_MAP,
    _adopt_pip_core_pattern,
    _pypi_to_conda_override,
    _resolve_adopt_pip_target,
)


class TestAdoptPipResolve(unittest.TestCase):
    def test_pypi_to_conda_override_map_is_normalized(self):
        # Keys should be directly usable with normalize_name().
        self.assertEqual(PYPI_TO_CONDA_NAME_MAP.get("opencv-python"), "opencv")
        self.assertEqual(PYPI_TO_CONDA_NAME_MAP.get("python-levenshtein"), "levenshtein")

    def test_pypi_to_conda_override_is_case_insensitive(self):
        self.assertEqual(_pypi_to_conda_override("Build"), "python-build")
        self.assertEqual(_pypi_to_conda_override("opencv-PYTHON"), "opencv")

    def test_core_pattern_replaces_separators(self):
        self.assertEqual(_adopt_pip_core_pattern("langchain_community"), "langchain*community")
        self.assertEqual(_adopt_pip_core_pattern("langchain-community"), "langchain*community")
        self.assertEqual(_adopt_pip_core_pattern("Flask-Cors"), "Flask*Cors")

    def test_core_pattern_does_not_inject_wildcards(self):
        self.assertEqual(_adopt_pip_core_pattern("pysha3"), "pysha3")
        self.assertEqual(_adopt_pip_core_pattern("0xcontractaddresses"), "0xcontractaddresses")

    def test_pass1_prefers_core_pattern_over_exact_name(self):
        # Mirrors adopt-pip pass1 term generation policy: if core differs, skip exact.
        pip_name = "streamlit-chat"
        core = _adopt_pip_core_pattern(pip_name)
        self.assertEqual(core, "streamlit*chat")
        terms = [core] if core != pip_name else [pip_name]
        self.assertEqual(terms, ["streamlit*chat"])

        pip_name2 = "pysha3"
        core2 = _adopt_pip_core_pattern(pip_name2)
        self.assertEqual(core2, "pysha3")
        terms2 = [core2] if core2 != pip_name2 else [pip_name2]
        self.assertEqual(terms2, ["pysha3"])

    def test_resolves_by_alnum_match(self):
        available = {"langchain-community", "something-else"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="langchain_community", available=available), "langchain-community")

    def test_resolves_msgpack_python_suffix(self):
        available = {"msgpack-python", "other"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="msgpack", available=available), "msgpack-python")

    def test_resolves_python_prefix_alias(self):
        available = {"python-build", "other"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="build", available=available), "python-build")

    def test_resolves_py_suffix_alias(self):
        available = {"ccxt-py", "other"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="ccxt", available=available), "ccxt-py")

    def test_resolves_safe_prefix_alias(self):
        available = {"safe-pysha3", "other"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="pysha3", available=available), "safe-pysha3")

    def test_prefers_exact_match(self):
        available = {"foo-bar", "foo_bar"}
        self.assertEqual(_resolve_adopt_pip_target(pip_name="foo_bar", available=available), "foo_bar")


if __name__ == "__main__":
    unittest.main()



import json
import os
import time
import sys
from pathlib import Path

from .conda_config import ensure_default_channels, load_conda_channels, load_pinned_specs
from .conda_ops import (
    conda_install,
    env_update_from_yaml,
    env_create_from_yaml,
    dry_run_install,
    clean_index_cache,
    conda_clean,
    conda_info_json,
    extract_pkgs_dirs,
    clean_cache_level,
    export_env_yaml,
    get_env_package_entries,
    is_conda_env,
    list_revisions,
    rollback_to_revision,
)
from .conflicts import find_same_version_case_conflicts
from .discovery import (
    detect_managers,
    discover_envs,
    env_name_from_path,
    get_python_exe,
    get_site_packages,
    select_envs,
    which,
)
from .i18n import t
from .naming import build_search_variants, normalize_name
from .pip_ops import pip_freeze, pip_install_requirements, pip_list_json, pip_reinstall, pip_uninstall
from .progress import Progress
from .scan import (
    remove_dist_info_paths,
    remove_invalid_artifact,
    scan_conda_meta_json,
    scan_dist_info,
    scan_invalid_artifacts,
    scan_pyd_duplicates,
)
from .search_parse import parse_search_output
from .subprocess_utils import OperationInterrupted, run_json_cmd
from .subprocess_utils import run_cmd_capture

from .clobber import build_conda_file_owner_map, extract_paths_from_text, to_relpath
from .inconsistent import parse_inconsistent
from .repair import (
    _adopt_pip,
    _apply_same_version_case_conflicts,
    _cleanup_duplicate_dist_info,
    _fix_conda_meta_issues,
    _fix_duplicates,
    _remove_invalid_artifacts,
)

def _debug(enabled, event, payload):
    if not enabled:
        return
    print("[debug]", f"{event}:", payload)




def scan_env(env_path):
    env = {"path": env_path, "python": None, "issues": []}
    python_exe = get_python_exe(env_path)
    if not python_exe:
        env["issues"].append({"type": "missing-python"})
        return env

    env["python"] = python_exe
    site_pkgs = get_site_packages(python_exe)
    if not site_pkgs:
        env["issues"].append({"type": "missing-site-packages"})
        return env

    for site_pkg in site_pkgs:
        if not Path(site_pkg).exists():
            continue
        env["issues"].extend(scan_dist_info(site_pkg))
        env["issues"].extend(scan_pyd_duplicates(site_pkg))
        env["issues"].extend(scan_invalid_artifacts(site_pkg))

    if is_conda_env(env_path):
        env["issues"].extend(scan_conda_meta_json(env_path))

    return env




def _print_fix_report(fixes, *, lang):
    if not fixes:
        print(t("fix_report", lang=lang))
        print(t("fix_report_none", lang=lang))
        return

    rows = []
    for f in fixes:
        action = f.get("method", "")
        item = f.get("package") or f.get("artifact") or ""
        if item == "<pip-to-conda>":
            item = "pip-to-conda({})".format(f.get("count", 0))
        result = t("fix_ok", lang=lang) if f.get("fixed") else t("fix_failed", lang=lang)
        reason_key = f.get("reason_key")
        if reason_key:
            reason = t(reason_key, lang=lang, **(f.get("reason_args") or {}))
        else:
            reason = ""
        rows.append((str(action), str(item), str(result), str(reason)))

    headers = (
        t("col_action", lang=lang),
        t("col_item", lang=lang),
        t("col_result", lang=lang),
        t("col_reason", lang=lang),
    )
    widths = [len(h) for h in headers]
    for r in rows:
        for i, cell in enumerate(r):
            widths[i] = max(widths[i], len(cell))

    def fmt_row(r):
        return "  ".join(str(r[i]).ljust(widths[i]) for i in range(4))

    print(t("fix_report", lang=lang))
    print(fmt_row(headers))
    print(fmt_row(("-" * widths[0], "-" * widths[1], "-" * widths[2], "-" * widths[3])))
    for r in rows:
        print(fmt_row(r))


def _approve(*, yes, plan, prompt, lang):
    if plan:
        print(t("plan_only", lang=lang))
        return False
    if yes:
        return True
    if not sys.stdin.isatty():
        return False
    try:
        ans = input(prompt).strip().lower()
    except KeyboardInterrupt:
        ans = ""
    return ans in ("y", "yes", "j", "ja")


def rollback(args):
    """
    Rollback command handler: snapshot -> conda revision rollback -> rescan.
    """
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    managers = detect_managers()
    has_any = any(managers.values())
    if not has_any:
        report = {"ok": False, "exit_code": 2, "error": t("manager_missing", lang=lang)}
        if args.json:
            return report
        print(t("manager_missing", lang=lang))
        return report

    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    if not targets:
        report = {"ok": False, "exit_code": 2, "error": "no target env"}
        return report
    env_path = targets[0]

    revs = list_revisions(env_path)
    if not revs:
        report = {"ok": False, "exit_code": 2, "error": t("rollback_no_revisions", lang=lang)}
        if not args.json:
            print(t("rollback_no_revisions", lang=lang))
        return report

    current = max(revs)
    to = args.to
    if to in (None, "", "prev"):
        target = current - 1
    elif to == "latest":
        target = current
    else:
        try:
            target = int(to)
        except Exception:
            target = None
    if target is None or target not in revs:
        report = {"ok": False, "exit_code": 2, "error": t("rollback_invalid_target", lang=lang, to=to)}
        if not args.json:
            print(t("rollback_invalid_target", lang=lang, to=to))
        return report

    if not args.json:
        print(t("rollback_plan", lang=lang, from_rev=current, to_rev=target))
    if getattr(args, "plan", False):
        if not args.json:
            print(t("plan_only", lang=lang))
        return {"ok": True, "exit_code": 0, "planned": True}

    if not _approve(yes=bool(args.yes), plan=False, prompt=t("prompt_approve", lang=lang), lang=lang):
        if not args.json:
            print(t("abort", lang=lang))
        return {"ok": False, "exit_code": 2, "aborted": True}

    # Snapshot (reuse existing mechanism by forcing a snapshot path)
    ts = time.strftime("%Y%m%d-%H%M%S")
    snap = Path(".env_repair") / "snapshots" / f"{env_name_from_path(env_path)}-{ts}" / "env.yml"
    snap.parent.mkdir(parents=True, exist_ok=True)
    snap_ok = False
    if is_conda_env(env_path) and manager:
        snap_ok = export_env_yaml(env_path, manager, snap)

    ok = rollback_to_revision(env_path, target, dry_run=bool(args.dry_run))
    if not args.json:
        if ok:
            print(t("rollback_done", lang=lang, to_rev=target))
    post = scan_env(env_path)
    post["snapshot"] = {"path": str(snap), "ok": snap_ok, "type": "conda-yaml"}
    post["action"] = {"type": "rollback", "from": current, "to": target, "ok": ok}
    exit_code = 0 if ok else 2
    return {"ok": ok, "exit_code": exit_code, "report": [post]}


def rebuild(args):
    """
    Rebuild command handler: export -> create new env -> optional verify scan.
    """
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    managers = detect_managers()
    has_any = any(managers.values())
    if not has_any:
        report = {"ok": False, "exit_code": 2, "error": t("manager_missing", lang=lang)}
        if not args.json:
            print(t("manager_missing", lang=lang))
        return report

    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    src_targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    if not src_targets:
        return {"ok": False, "exit_code": 2, "error": "no source env"}
    src = src_targets[0]

    to = args.to
    target_is_path = (":" in to) or ("\\" in to) or ("/" in to)
    dst = Path(to).resolve() if target_is_path else to

    # Existence check
    if target_is_path:
        if Path(dst).exists():
            msg = t("rebuild_target_exists", lang=lang, dst=str(dst))
            if not args.json:
                print(msg)
            return {"ok": False, "exit_code": 2, "error": msg}
    else:
        # name: avoid clobbering an existing env with same name
        existing = {Path(p).name.lower() for p in all_envs}
        if str(dst).lower() in existing:
            msg = t("rebuild_target_exists", lang=lang, dst=str(dst))
            if not args.json:
                print(msg)
            return {"ok": False, "exit_code": 2, "error": msg}

    if not args.json:
        print(t("rebuild_plan", lang=lang, src=src, dst=str(dst)))
        if getattr(args, "plan", False):
            print(t("plan_only", lang=lang))
            return {"ok": True, "exit_code": 0, "planned": True}

    if not _approve(yes=bool(args.yes), plan=False, prompt=t("prompt_approve", lang=lang), lang=lang):
        if not args.json:
            print(t("abort", lang=lang))
        return {"ok": False, "exit_code": 2, "aborted": True}

    ts = time.strftime("%Y%m%d-%H%M%S")
    snap = Path(".env_repair") / "snapshots" / f"{env_name_from_path(src)}-{ts}" / "env.yml"
    snap.parent.mkdir(parents=True, exist_ok=True)
    snap_ok = False
    if is_conda_env(src) and manager:
        snap_ok = export_env_yaml(src, manager, snap)

    created = env_create_from_yaml(manager=manager, src_yaml=snap, target=dst, target_is_path=target_is_path)

    result_report = {"path": str(dst), "python": None, "issues": []}
    if args.verify and created:
        # If name-based, discover target path and scan it.
        if not target_is_path:
            all2, base2, _mgr2 = discover_envs(show_json_output=show_json_output)
            tgt = select_envs(all2, [str(dst)], base2)
            if tgt:
                result_report = scan_env(tgt[0])
        else:
            result_report = scan_env(str(dst))

    out = {
        "ok": bool(created),
        "exit_code": 0 if created else 2,
        "report": [
            {
                "source": src,
                "snapshot": {"path": str(snap), "ok": snap_ok, "type": "conda-yaml"},
                "action": {"type": "rebuild", "to": str(dst), "ok": bool(created), "verify": bool(args.verify)},
                "result": result_report,
            }
        ],
    }
    if not args.json and created:
        print(t("rebuild_done", lang=lang, dst=str(dst)))
    return out


def diagnose_clobber(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    all_envs, base_prefix, _manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    if not targets:
        return {"ok": False, "exit_code": 2, "error": "no target env"}
    env_path = targets[0]

    if not args.logfile:
        msg = t("clobber_no_log", lang=lang)
        if not args.json:
            print(t("clobber_header", lang=lang))
            print(msg)
        return {"ok": False, "exit_code": 2, "error": msg, "report": []}

    log_path = Path(args.logfile)
    try:
        text = log_path.read_text(encoding="utf-8", errors="ignore")
    except OSError:
        msg = t("clobber_log_read_failed", lang=lang, path=str(log_path))
        if not args.json:
            print(t("clobber_header", lang=lang))
            print(msg)
        return {"ok": False, "exit_code": 2, "error": msg, "report": []}

    paths = extract_paths_from_text(text, env_prefix=env_path)
    owners = build_conda_file_owner_map(env_path)

    conflicts = []
    for p in paths:
        rel = to_relpath(env_path, p)
        owner = owners.get(rel) if rel else None
        conflicts.append({"path": p, "relpath": rel, "conda_owner": owner})

    ok = bool(conflicts)
    if not args.json:
        print(t("clobber_header", lang=lang))
        if not conflicts:
            print(t("clobber_no_paths", lang=lang))

    return {"ok": ok, "exit_code": 0 if ok else 1, "report": [{"env": env_path, "conflicts": conflicts}]}


def diagnose_inconsistent(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    all_envs, base_prefix, _manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    if not targets:
        return {"ok": False, "exit_code": 2, "error": "no target env"}
    env_path = targets[0]

    rc, out, err = dry_run_install(env_path, ["python"])
    inconsistent, pkgs = parse_inconsistent(out + "\n" + err)
    if not args.json:
        print(t("inconsistent_header", lang=lang))
        print(t("inconsistent_found", lang=lang) if inconsistent else t("inconsistent_not_found", lang=lang))
    return {
        "ok": True,
        "exit_code": 0,
        "report": [{"env": env_path, "dry_run_rc": rc, "inconsistent": inconsistent, "packages": pkgs}],
    }


def fix_inconsistent(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    if not targets:
        return {"ok": False, "exit_code": 2, "error": "no target env"}
    env_path = targets[0]

    level = args.level
    if not args.json:
        print(t("inconsistent_fix_plan", lang=lang, level=level, env=env_path))
        if getattr(args, "plan", False):
            print(t("plan_only", lang=lang))
            return {"ok": True, "exit_code": 0, "planned": True}

    if not _approve(yes=bool(args.yes), plan=False, prompt=t("prompt_approve", lang=lang), lang=lang):
        if not args.json:
            print(t("abort", lang=lang))
        return {"ok": False, "exit_code": 2, "aborted": True}

    actions = []
    ok = True

    if level in ("safe", "normal"):
        ok = clean_index_cache(yes=True)
        actions.append({"type": "clean_index_cache", "ok": ok})
        if ok and level == "normal":
            # Best-effort: reinstall packages listed in warning, if any.
            rc, out, err = dry_run_install(env_path, ["python"])
            inconsistent, pkgs = parse_inconsistent(out + "\n" + err)
            if inconsistent and pkgs:
                ok2 = conda_install(
                    env_path,
                    pkgs,
                    manager,
                    [],
                    ignore_pinned=False,
                    force_reinstall=True,
                )
                actions.append({"type": "force_reinstall", "packages": pkgs, "ok": ok2})
                ok = ok and ok2
    elif level == "rebuild":
        # Non-invasive: suggest using rebuild.
        actions.append({"type": "suggest_rebuild", "ok": True})
        ok = True
    else:
        return {"ok": False, "exit_code": 2, "error": "invalid level"}

    post = scan_env(env_path)
    if not args.json:
        print(t("inconsistent_fix_done", lang=lang, level=level))
    return {"ok": ok, "exit_code": 0 if ok else 1, "report": [{"env": env_path, "actions": actions, "post": post}]}


def cache_check(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"
    info = conda_info_json(show_json_output=show_json_output)
    pkgs_dirs = extract_pkgs_dirs(info)
    if not args.json:
        print(t("cache_header", lang=lang))
        print(t("cache_pkgs_dirs", lang=lang, value=", ".join(pkgs_dirs) if pkgs_dirs else "-"))
    return {"ok": True, "exit_code": 0, "report": [{"pkgs_dirs": pkgs_dirs}]}


def cache_fix(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"
    level = args.level
    steps = clean_cache_level(level)
    if not steps:
        return {"ok": False, "exit_code": 2, "error": "invalid level"}

    if not args.json:
        print(t("cache_fix_plan", lang=lang, level=level))
        if getattr(args, "plan", False):
            print(t("plan_only", lang=lang))
            return {"ok": True, "exit_code": 0, "planned": True}

    if not _approve(yes=bool(args.yes), plan=False, prompt=t("prompt_approve", lang=lang), lang=lang):
        if not args.json:
            print(t("abort", lang=lang))
        return {"ok": False, "exit_code": 2, "aborted": True}

    actions = []
    ok = True
    for step in steps:
        step_ok = conda_clean(step, yes=True)
        actions.append({"type": "conda_clean", "args": step, "ok": step_ok})
        ok = ok and step_ok

    info = conda_info_json(show_json_output=show_json_output)
    pkgs_dirs = extract_pkgs_dirs(info)
    if not args.json:
        print(t("cache_fix_done", lang=lang, level=level))
    return {"ok": ok, "exit_code": 0 if ok else 1, "report": [{"level": level, "actions": actions, "pkgs_dirs": pkgs_dirs}]}


def diagnose_ssl(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    if getattr(args, "base", False):
        env_path = base_prefix
    else:
        targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
        env_path = targets[0] if targets else None

    if not env_path:
        return {"ok": False, "exit_code": 2, "error": "no target env"}

    python_exe = get_python_exe(env_path)
    info = conda_info_json(show_json_output=show_json_output)

    ssl_out = ""
    ssl_err = ""
    ssl_rc = 1
    if python_exe:
        ssl_rc, ssl_out, ssl_err = run_cmd_capture([python_exe, "-c", "import ssl; print(ssl.OPENSSL_VERSION)"])
    ok = ssl_rc == 0
    if not args.json:
        print(t("ssl_header", lang=lang))
        if ok:
            print(t("ssl_ok", lang=lang, value=ssl_out.strip()))
        else:
            msg = ssl_err.strip() or ssl_out.strip() or str(ssl_rc)
            print(t("ssl_fail", lang=lang, value=msg))
            print(t("ssl_hint", lang=lang))

    return {
        "ok": True,
        "exit_code": 0,
        "report": [
            {
                "env": env_path,
                "python": python_exe,
                "ssl": {"ok": ok, "rc": ssl_rc, "stdout": ssl_out, "stderr": ssl_err},
                "conda_info": {"pkgs_dirs": extract_pkgs_dirs(info)},
                "manager": manager,
            }
        ],
    }


def run(args):
    show_json_output = bool(getattr(args, "debug", False))
    lang = "auto"

    managers = detect_managers()
    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, args.env, base_prefix)

    pip_fallback = bool(args.pip_fallback)
    if args.no_pip_fallback:
        pip_fallback = False

    channels = []
    if not args.no_channels_from_condarc:
        channels.extend(
            load_conda_channels(base_prefix=base_prefix, has_conda=which("conda") or which("mamba"), show_json_output=show_json_output)
        )
    if args.channel:
        channels.extend(args.channel)
    if not args.no_default_channels:
        channels = ensure_default_channels(channels)

    report = []
    ok_all = True
    exit_code = 0

    env_progress = None
    if not args.json and targets:
        env_progress = Progress(total=len(targets), label=t("progress_envs", lang=lang))

    for env_idx, env_path in enumerate(targets, start=1):
        if env_progress:
            env_progress.update(env_idx)
        if not args.json:
            print(t("step_scan", lang=lang) + ": " + env_path)
        env_report = scan_env(env_path)
        env_report["managers"] = {
            "conda": {"found": bool(managers.get("conda")), "path": managers.get("conda")},
            "mamba": {"found": bool(managers.get("mamba")), "path": managers.get("mamba")},
            "micromamba": {"found": bool(managers.get("micromamba")), "path": managers.get("micromamba")},
        }
        env_report["channels"] = list(channels)
        env_report["pinned"] = load_pinned_specs(env_path)

        python_exe = env_report.get("python")
        conda_here = bool(manager) and is_conda_env(env_path)
        if conda_here:
            entries = get_env_package_entries(env_path, manager, show_json_output=show_json_output)
        else:
            entries = pip_list_json(python_exe) if python_exe else []
        env_report["initial_entries"] = entries

        snapshot = None
        if args.snapshot:
            snapshot = Path(args.snapshot)
        elif args.fix:
            # Always create a rescue snapshot before modifications.
            ts = time.strftime("%Y%m%d-%H%M%S")
            base = Path(".env_repair") / "snapshots"
            name = env_name_from_path(env_path)
            snapshot = base / f"{name}-{ts}" / ("env.yml" if conda_here else "requirements.txt")

        if snapshot:
            if conda_here and manager:
                if not args.json:
                    print(t("step_snapshot", lang=lang) + ": " + env_path)
                snap_ok = export_env_yaml(env_path, manager, snapshot)
                env_report["snapshot"] = {"path": str(snapshot), "ok": snap_ok, "type": "conda-yaml"}
            elif python_exe:
                if not args.json:
                    print(t("step_snapshot", lang=lang) + ": " + env_path)
                snap_ok = pip_freeze(python_exe, snapshot)
                env_report["snapshot"] = {"path": str(snapshot), "ok": snap_ok, "type": "pip-freeze"}
            else:
                env_report["snapshot"] = {"path": str(snapshot), "ok": False, "reason": "no-python"}

        if args.fix:
            fixes = []
            try:
                fixes.extend(_remove_invalid_artifacts(env_report, args.debug))
                if conda_here:
                    fixes.extend(
                        _fix_conda_meta_issues(
                            env_report,
                            manager,
                            channels,
                            args.ignore_pinned,
                            args.force_reinstall,
                            args.debug,
                        )
                    )
                fixes.extend(_cleanup_duplicate_dist_info(env_report, args.debug))
                fixes.extend(
                    _apply_same_version_case_conflicts(
                        env_report,
                        entries,
                        manager,
                        channels,
                        args.ignore_pinned,
                        args.force_reinstall,
                        args.debug,
                    )
                    )
                fixes.extend(
                    _fix_duplicates(
                        env_report,
                        entries,
                        manager if conda_here else None,
                        channels,
                        args.ignore_pinned,
                        args.force_reinstall,
                        args.prefer,
                        pip_fallback,
                        args.debug,
                        in_conda_env=bool(conda_here),
                    )
                )

                if args.adopt_pip and conda_here:
                    if not args.json:
                        print(t("step_adopt_pip", lang=lang) + ": " + env_path)
                    fixes.extend(
                        _adopt_pip(
                            env_report,
                            entries,
                            manager,
                            channels,
                            args.ignore_pinned,
                            args.force_reinstall,
                            not args.keep_pip,
                            args.debug,
                            show_json_output=show_json_output,
                            lang=lang,
                        )
                    )
            except OperationInterrupted as e:
                env_report["interrupted"] = {
                    "cmd": e.cmd,
                    "returncode": e.returncode,
                    "snapshot": env_report.get("snapshot"),
                }
                exit_code = max(exit_code, e.returncode)
                ok_all = False
                fixes.append({"fixed": False, "method": "interrupted", "package": "<operation>"})

                state_path = Path(".env_repair") / "state.json"
                state_path.parent.mkdir(parents=True, exist_ok=True)
                try:
                    state_path.write_text(
                        json.dumps(
                            {
                                "env_path": env_path,
                                "snapshot": env_report.get("snapshot"),
                                "cmd": e.cmd,
                                "when": time.strftime("%Y-%m-%d %H:%M:%S"),
                            },
                            indent=2,
                        ),
                        encoding="utf-8",
                    )
                except OSError:
                    pass

                if args.json or not sys.stdin.isatty():
                    env_report["fixes"] = fixes
                    report.append(env_report)
                    break
                try:
                    choice = input(t("prompt_interrupted", lang=lang)).strip().lower()
                except KeyboardInterrupt:
                    choice = "a"
                if choice == "r":
                    snap = env_report.get("snapshot") or {}
                    snap_path = snap.get("path")
                    restored = False
                    if snap_path and snap.get("ok"):
                        if conda_here and manager and snap.get("type") == "conda-yaml":
                            restored = env_update_from_yaml(env_path, manager, snap_path)
                        elif python_exe and snap.get("type") == "pip-freeze":
                            restored = pip_install_requirements(python_exe, snap_path)
                    fixes.append({"fixed": restored, "method": "restore", "package": "<snapshot>"})
                    if not restored:
                        exit_code = max(exit_code, 1)
                        break
                elif choice == "c":
                    pass
                else:
                    break

            env_report["fixes"] = fixes
            ok_all = ok_all and all(f.get("fixed") for f in fixes)

        report.append(env_report)

    if not args.json:
        if env_progress:
            env_progress.finish()
        for env in report:
            name = env_name_from_path(env["path"])
            print(t("env_header", lang=lang, name=name))
            print(t("path", lang=lang, value=env["path"]))
            print(t("python", lang=lang, value=env.get("python") or "missing"))
            mgrs = env.get("managers") or {}
            if mgrs:
                # Keep this short; details are in JSON output.
                missing = [k for k, v in mgrs.items() if not (v or {}).get("found")]
                if len(missing) == 3:
                    print(t("manager_missing", lang=lang))
            if env.get("channels"):
                print(t("channels", lang=lang, value=", ".join(env["channels"])))
            if env.get("snapshot"):
                s = env["snapshot"]
                status = t("snapshot_ok", lang=lang) if s.get("ok") else t("snapshot_failed", lang=lang)
                print(t("snapshot", lang=lang, path=s.get("path"), status=status))
            if not env.get("issues"):
                print(t("issues_none", lang=lang))
            else:
                print(t("issues", lang=lang))
                for issue in env["issues"]:
                    issue_type = issue.get("type")
                    if issue_type == "duplicate-dist-info":
                        print(
                            t(
                                "issue_duplicate_dist_info",
                                lang=lang,
                                package=issue.get("package"),
                                versions=issue.get("versions"),
                            )
                        )
                    elif issue_type == "duplicate-pyd":
                        print(t("issue_duplicate_pyd", lang=lang, base=issue.get("base"), files=issue.get("files")))
                    elif issue_type == "invalid-artifact":
                        print(t("issue_invalid_artifact", lang=lang, name=issue.get("name")))
                    else:
                        print(t("issue_generic", lang=lang, type=issue_type))
            if env.get("pinned"):
                print(t("pinned", lang=lang))
                for p in env["pinned"]:
                    print(" -", p)
            if args.fix:
                fixes = env.get("fixes", [])
                if not fixes:
                    print(t("fixes_none", lang=lang))
                else:
                    print(t("fixes", lang=lang))
                    for f in fixes:
                        method = f.get("method", "unknown")
                        label = f.get("package") or f.get("artifact") or "item"
                        if label == "<pip-to-conda>":
                            label = "pip-to-conda({})".format(f.get("count", 0))
                        status = t("fix_ok", lang=lang) if f.get("fixed") else t("fix_failed", lang=lang)
                        print(t("fix_line", lang=lang, label=label, method=method, status=status))
                    _print_fix_report(fixes, lang=lang)
            print("")

    if exit_code == 0 and not ok_all:
        exit_code = 1
    return {"ok": ok_all, "report": report, "exit_code": exit_code}



def collect_name_fields(obj, names):
    if isinstance(obj, dict):
        for k, v in obj.items():
            if k == "name" and isinstance(v, str):
                names.add(v)
            collect_name_fields(v, names)
    elif isinstance(obj, list):
        for item in obj:
            collect_name_fields(item, names)


def extract_search_results(data):
    if not isinstance(data, dict):
        return []
    if isinstance(data.get("result"), dict):
        result = data["result"]
        pkgs = result.get("pkgs")
        if isinstance(pkgs, list):
            names = []
            for item in pkgs:
                if isinstance(item, dict) and isinstance(item.get("name"), str):
                    names.append(item["name"])
            return names
        if all(isinstance(v, list) for v in result.values()):
            return list(result.keys())
        names = []
        for v in result.values():
            if isinstance(v, list):
                for item in v:
                    if isinstance(item, dict) and isinstance(item.get("name"), str):
                        names.append(item["name"])
        return names
    if isinstance(data.get("pkgs"), dict):
        return list(data["pkgs"].keys())
    if isinstance(data.get("packages"), dict):
        return list(data["packages"].keys())
    if isinstance(data.get("result"), list):
        names = []
        for item in data["result"]:
            if isinstance(item, dict):
                name = item.get("name")
                if name:
                    names.append(name)
                    continue
                rec = item.get("record") if isinstance(item.get("record"), dict) else None
                if rec and isinstance(rec.get("name"), str):
                    names.append(rec["name"])
                continue
            if isinstance(item, str):
                names.append(item)
        return names
    if all(isinstance(v, list) for v in data.values()):
        return list(data.keys())
    return []


def parse_search_output(data):
    results = extract_search_results(data)
    if results:
        return results
    names = set()
    collect_name_fields(data, names)
    return sorted(names)




import unittest

import env_repair as ed


class TestSearchJsonParsing(unittest.TestCase):
    def test_extract_search_results_wildcard_result_pkgs(self):
        data = {
            "query": {"query": "msgpack*", "type": "search"},
            "result": {
                "msg": "",
                "pkgs": [
                    {"name": "msgpack-numpy", "version": "0.4.8"},
                    {"name": "msgpack-python", "version": "1.1.2"},
                ],
                "status": "OK",
            },
        }
        self.assertEqual(set(ed.extract_search_results(data)), {"msgpack-numpy", "msgpack-python"})

    def test_extract_search_results_result_dict_keys(self):
        data = {
            "result": {
                "pyqt": [{"name": "pyqt", "version": "5.15.11"}],
                "zope.interface": [{"name": "zope.interface", "version": "8.0.1"}],
            }
        }
        self.assertEqual(set(ed.extract_search_results(data)), {"pyqt", "zope.interface"})


if __name__ == "__main__":
    unittest.main()



import os
import re
from pathlib import Path

from .subprocess_utils import run_json_cmd


def load_conda_channels_from_condarc(*, base_prefix=None):
    paths = []
    env_condarc = os.environ.get("CONDARC")
    if env_condarc:
        paths.append(Path(env_condarc))
    paths.append(Path.home() / ".condarc")
    if base_prefix:
        paths.append(Path(base_prefix) / ".condarc")

    for p in paths:
        if not p.exists():
            continue
        try:
            lines = p.read_text(encoding="utf-8", errors="ignore").splitlines()
        except OSError:
            continue
        in_channels = False
        channels = []
        for line in lines:
            raw = line.strip()
            if not raw or raw.startswith("#"):
                continue
            if raw.startswith("channels:"):
                in_channels = True
                continue
            if in_channels:
                if raw.startswith("- "):
                    channels.append(raw[2:].strip())
                elif re.match(r"^[A-Za-z0-9_-]+\\s*:", raw):
                    break
        if channels:
            return channels
    return []


def load_conda_channels(*, base_prefix=None, has_conda, show_json_output):
    channels = load_conda_channels_from_condarc(base_prefix=base_prefix)
    if channels:
        return channels
    if has_conda:
        data = run_json_cmd(["conda", "config", "--show", "channels", "--json"], show_json_output=show_json_output)
        if data:
            channels = data.get("channels") or []
            return [c for c in channels if isinstance(c, str)]
    return []


def ensure_default_channels(channels):
    if "defaults" not in channels:
        channels.append("defaults")
    if "anaconda" not in channels:
        channels.append("anaconda")
    return channels


def load_pinned_specs(env_path):
    pinned = []
    pinned_file = Path(env_path, "conda-meta", "pinned")
    if pinned_file.exists():
        for line in pinned_file.read_text(encoding="utf-8", errors="ignore").splitlines():
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            pinned.append(line)
    return sorted(set(pinned))




import json
import tempfile
import unittest
from pathlib import Path

from env_repair.clobber import build_conda_file_owner_map, extract_paths_from_text, to_relpath


class TestClobberParse(unittest.TestCase):
    def test_extracts_paths_inside_prefix(self):
        prefix = r"C:\Anaconda3"
        log = r"ClobberError: The package cannot be installed due to a path conflict.\n  path: 'C:\\Anaconda3\\Lib\\site-packages\\x.py'\n"
        paths = extract_paths_from_text(log, env_prefix=prefix)
        self.assertEqual(paths, [r"C:\Anaconda3\Lib\site-packages\x.py"])

    def test_owner_map_from_conda_meta_files(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            cm = env / "conda-meta"
            cm.mkdir(parents=True, exist_ok=True)
            record = cm / "demo-1.0-0.json"
            record.write_text(
                json.dumps({"name": "demo", "version": "1.0", "build": "0", "files": ["Lib/site-packages/x.py"]}),
                encoding="utf-8",
            )
            owners = build_conda_file_owner_map(str(env))
            self.assertIn("Lib/site-packages/x.py", owners)
            self.assertEqual(owners["Lib/site-packages/x.py"]["name"], "demo")

    def test_relpath(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            p = env / "Lib" / "site-packages" / "x.py"
            p.parent.mkdir(parents=True, exist_ok=True)
            p.write_text("x", encoding="utf-8")
            rel = to_relpath(str(env), str(p))
            self.assertEqual(rel, "Lib/site-packages/x.py")


if __name__ == "__main__":
    unittest.main()




import json
import tempfile
import unittest
from pathlib import Path

from env_repair.scan import scan_conda_meta_json


class TestScanCondaMetaJson(unittest.TestCase):
    def test_missing_depends_not_flagged_if_record_complete(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            meta = env / "conda-meta"
            meta.mkdir(parents=True, exist_ok=True)
            p = meta / "pybind11-abi-11-hc364b38_1.json"
            p.write_text(json.dumps({"name": "pybind11-abi", "version": "11", "build": "hc364b38_1"}), encoding="utf-8")
            issues = scan_conda_meta_json(str(env))
            self.assertEqual(issues, [])

    def test_missing_depends_flagged_if_record_incomplete(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            meta = env / "conda-meta"
            meta.mkdir(parents=True, exist_ok=True)
            p = meta / "somepkg-1.0-0.json"
            p.write_text(json.dumps({"version": "1.0"}), encoding="utf-8")
            issues = scan_conda_meta_json(str(env))
            self.assertEqual(len(issues), 1)
            self.assertEqual(issues[0].get("type"), "conda-meta-missing-depends")


if __name__ == "__main__":
    unittest.main()




import json
import os
import subprocess
import sys
import threading


class OperationInterrupted(RuntimeError):
    def __init__(self, cmd, *, returncode=130):
        super().__init__("Operation interrupted")
        self.cmd = cmd
        self.returncode = returncode


def _as_cmd_exe(cmd):
    """
    Windows fallback: execute via cmd.exe so that .bat/.cmd shims in PATH work.
    """
    if not cmd:
        return cmd
    # list2cmdline is Windows-aware quoting.
    return ["cmd", "/d", "/c", subprocess.list2cmdline(cmd)]


def _cmd_name(cmd):
    if not cmd:
        return ""
    first = cmd[0]
    if not isinstance(first, str):
        return ""
    base = os.path.basename(first).lower()
    if base.endswith(".exe"):
        base = base[:-4]
    return base


def _should_print_cmd(cmd):
    # Enabled via CLI `--debug` (sets ENV var) or explicit env var usage.
    if os.environ.get("ENV_REPAIR_SHOW_CMDS") not in ("1", "true", "TRUE", "yes", "YES"):
        return False
    # User asked specifically for conda/mamba commandlines.
    return _cmd_name(cmd) in {"conda", "mamba", "micromamba"}


def _print_cmd(cmd):
    try:
        rendered = subprocess.list2cmdline(cmd)
    except Exception:
        rendered = str(cmd)
    sys.stderr.write(f"[cmd] {rendered}\n")
    sys.stderr.flush()


def run_cmd_capture(cmd):
    """
    Run a command and capture stdout/stderr.
    Returns CompletedProcess-like tuple: (returncode, stdout, stderr).
    """
    if _should_print_cmd(cmd):
        _print_cmd(cmd)
    try:
        res = subprocess.run(cmd, capture_output=True, text=True, check=False)
    except FileNotFoundError:
        if os.name == "nt":
            res = subprocess.run(_as_cmd_exe(cmd), capture_output=True, text=True, check=False)
        else:
            raise
    except KeyboardInterrupt as e:
        raise OperationInterrupted(cmd, returncode=130) from e
    return res.returncode, res.stdout, res.stderr


def run_cmd_live(cmd):
    """
    Run a command and stream stdout/stderr live.

    Note: We intentionally pipe+stream instead of inheriting the parent console.
    For tools like (mamba/conda) this avoids fancy TUI/progress output that can be
    swallowed by some terminals, and guarantees we surface logs + warnings.
    """
    if _should_print_cmd(cmd):
        _print_cmd(cmd)
    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1)
    except FileNotFoundError:
        if os.name == "nt":
            proc = subprocess.Popen(
                _as_cmd_exe(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
            )
        else:
            raise
    out_buf = []
    err_buf = []
    t_out = threading.Thread(target=_stream_reader, args=(proc.stdout, sys.stdout, out_buf))
    t_err = threading.Thread(target=_stream_reader, args=(proc.stderr, sys.stderr, err_buf))
    t_out.start()
    t_err.start()
    try:
        rc = proc.wait()
    except KeyboardInterrupt as e:
        try:
            proc.terminate()
        except Exception:
            pass
        try:
            proc.wait(timeout=5)
        except Exception:
            pass
        raise OperationInterrupted(cmd, returncode=130) from e
    t_out.join()
    t_err.join()
    return rc


def run_cmd_live_capture(cmd):
    """
    Run a command, stream stdout/stderr live, and also return captured output.

    Returns: (returncode, stdout_text, stderr_text).
    """
    if _should_print_cmd(cmd):
        _print_cmd(cmd)
    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1)
    except FileNotFoundError:
        if os.name == "nt":
            proc = subprocess.Popen(
                _as_cmd_exe(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
            )
        else:
            raise
    out_buf = []
    err_buf = []
    t_out = threading.Thread(target=_stream_reader, args=(proc.stdout, sys.stdout, out_buf))
    t_err = threading.Thread(target=_stream_reader, args=(proc.stderr, sys.stderr, err_buf))
    t_out.start()
    t_err.start()
    try:
        rc = proc.wait()
    except KeyboardInterrupt as e:
        try:
            proc.terminate()
        except Exception:
            pass
        try:
            proc.wait(timeout=5)
        except Exception:
            pass
        raise OperationInterrupted(cmd, returncode=130) from e
    t_out.join()
    t_err.join()
    return rc, "".join(out_buf), "".join(err_buf)


def _stream_reader(stream, sink, buffer):
    for line in iter(stream.readline, ""):
        if sink:
            sink.write(line)
            sink.flush()
        if buffer is not None:
            buffer.append(line)


def run_json_cmd(cmd, *, show_json_output):
    if _should_print_cmd(cmd):
        _print_cmd(cmd)
    if not show_json_output:
        try:
            res = subprocess.run(cmd, capture_output=True, text=True, check=False)
        except FileNotFoundError:
            if os.name == "nt":
                res = subprocess.run(_as_cmd_exe(cmd), capture_output=True, text=True, check=False)
            else:
                raise
        except KeyboardInterrupt as e:
            raise OperationInterrupted(cmd, returncode=130) from e
        if res.returncode != 0:
            return None
        try:
            return json.loads(res.stdout)
        except ValueError:
            return None

    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, bufsize=1)
    except FileNotFoundError:
        if os.name == "nt":
            proc = subprocess.Popen(
                _as_cmd_exe(cmd),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
            )
        else:
            raise
    out_buf = []
    err_buf = []
    t_out = threading.Thread(target=_stream_reader, args=(proc.stdout, sys.stdout, out_buf))
    t_err = threading.Thread(target=_stream_reader, args=(proc.stderr, sys.stderr, err_buf))
    t_out.start()
    t_err.start()
    try:
        rc = proc.wait()
    except KeyboardInterrupt as e:
        try:
            proc.terminate()
        except Exception:
            pass
        try:
            proc.wait(timeout=5)
        except Exception:
            pass
        raise OperationInterrupted(cmd, returncode=130) from e
    t_out.join()
    t_err.join()
    if rc != 0:
        return None
    try:
        return json.loads("".join(out_buf))
    except ValueError:
        return None



import shutil
import subprocess
import sys
from pathlib import Path

from .subprocess_utils import run_cmd_capture, run_cmd_live, run_cmd_live_capture, run_json_cmd


def is_conda_env(env_path):
    return Path(env_path, "conda-meta").exists()


def ensure_mamba(*, base_prefix):
    import shutil

    if shutil.which("mamba"):
        return True
    if not shutil.which("conda") or not base_prefix:
        return False
    cmd = ["conda", "install", "-y", "-p", base_prefix, "mamba"]
    return subprocess.run(cmd, check=False).returncode == 0 and shutil.which("mamba")


def get_env_package_entries(env_path, manager, *, show_json_output):
    if manager == "micromamba":
        cmd = ["micromamba", "list", "-p", env_path, "--json"]
    elif manager == "mamba":
        cmd = ["mamba", "list", "-p", env_path, "--json"]
    else:
        cmd = ["conda", "list", "-p", env_path, "--json"]
    data = run_json_cmd(cmd, show_json_output=show_json_output)
    if not data or not isinstance(data, list):
        return []
    return [p for p in data if isinstance(p, dict)]


def conda_install(env_path, packages, manager, channels, *, ignore_pinned, force_reinstall):
    if not packages:
        return True
    channel_args = []
    for ch in channels or []:
        channel_args.extend(["-c", ch])
    pin_args = ["--no-pin"] if ignore_pinned else []
    force_args = ["--force-reinstall"] if force_reinstall else []
    if manager == "micromamba":
        cmd = ["micromamba", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    elif manager == "mamba":
        cmd = ["mamba", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    else:
        # Some conda versions do not support `--update-deps` (mamba does).
        cmd = ["conda", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    return run_cmd_live(cmd) == 0


def conda_install_capture(env_path, packages, manager, channels, *, ignore_pinned, force_reinstall):
    """
    Like `conda_install` but returns (ok, stdout, stderr) while still streaming output live.
    """
    if not packages:
        return True, "", ""
    channel_args = []
    for ch in channels or []:
        channel_args.extend(["-c", ch])
    pin_args = ["--no-pin"] if ignore_pinned else []
    force_args = ["--force-reinstall"] if force_reinstall else []
    if manager == "micromamba":
        cmd = ["micromamba", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    elif manager == "mamba":
        cmd = ["mamba", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    else:
        cmd = ["conda", "install", "-y", "-p", env_path] + force_args + pin_args + channel_args + list(packages)
    rc, out, err = run_cmd_live_capture(cmd)
    return rc == 0, out, err


def conda_remove(env_path, packages, manager):
    if not packages:
        return True
    if manager == "micromamba":
        cmd = ["micromamba", "remove", "-y", "-p", env_path] + list(packages)
    elif manager == "mamba":
        cmd = ["mamba", "remove", "-y", "-p", env_path] + list(packages)
    else:
        cmd = ["conda", "remove", "-y", "-p", env_path] + list(packages)
    return run_cmd_live(cmd) == 0


def export_env_yaml(env_path, manager, out_path):
    out_path = Path(out_path)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    if manager == "micromamba":
        cmd = ["micromamba", "env", "export", "-p", env_path]
    elif manager == "mamba":
        cmd = ["mamba", "env", "export", "-p", env_path]
    elif manager == "conda":
        cmd = ["conda", "env", "export", "-p", env_path]
    else:
        return False
    with out_path.open("w", encoding="utf-8") as f:
        proc = subprocess.Popen(cmd, stdout=f, stderr=sys.stderr)
        return proc.wait() == 0


def env_update_from_yaml(env_path, manager, yaml_path):
    if manager == "micromamba":
        cmd = ["micromamba", "env", "update", "-p", env_path, "-f", str(yaml_path)]
    elif manager == "mamba":
        cmd = ["mamba", "env", "update", "-p", env_path, "-f", str(yaml_path)]
    elif manager == "conda":
        cmd = ["conda", "env", "update", "-p", env_path, "-f", str(yaml_path)]
    else:
        return False
    return run_cmd_live(cmd) == 0


def list_revisions(env_path):
    """
    Returns a sorted list of available revision numbers for the env.
    Uses `conda list --revisions` (text output) and parses the leading revision column.
    """
    import re

    if shutil.which("conda"):
        cmd = ["conda", "list", "--revisions", "-p", env_path]
    elif shutil.which("mamba"):
        cmd = ["mamba", "list", "--revisions", "-p", env_path]
    else:
        return []

    rc, out, _err = run_cmd_capture(cmd)
    if rc != 0 or not out:
        return []
    revs = []
    for line in out.splitlines():
        m = re.match(r"^\\s*(\\d+)\\s+", line)
        if m:
            try:
                revs.append(int(m.group(1)))
            except ValueError:
                pass
    return sorted(set(revs))


def rollback_to_revision(env_path, revision, *, dry_run):
    """
    Rollback an environment to a specific conda revision number.
    Returns bool.
    """
    runner = "conda" if shutil.which("conda") else ("mamba" if shutil.which("mamba") else None)
    if not runner:
        return False
    cmd = [runner, "install", "-y", "-p", env_path, "--revision", str(int(revision))]
    if dry_run:
        cmd.insert(2, "--dry-run")
    return run_cmd_live(cmd) == 0


def env_create_from_yaml(*, manager, src_yaml, target, target_is_path):
    if manager == "micromamba":
        # micromamba supports env create -f and -p
        if target_is_path:
            cmd = ["micromamba", "env", "create", "-f", str(src_yaml), "-p", str(target)]
        else:
            cmd = ["micromamba", "env", "create", "-f", str(src_yaml), "-n", str(target)]
    elif manager == "mamba":
        if target_is_path:
            cmd = ["mamba", "env", "create", "-f", str(src_yaml), "-p", str(target)]
        else:
            cmd = ["mamba", "env", "create", "-f", str(src_yaml), "-n", str(target)]
    elif manager == "conda":
        if target_is_path:
            cmd = ["conda", "env", "create", "-f", str(src_yaml), "-p", str(target)]
        else:
            cmd = ["conda", "env", "create", "-f", str(src_yaml), "-n", str(target)]
    else:
        return False
    return run_cmd_live(cmd) == 0


def dry_run_install(env_path, packages):
    runner = "conda" if shutil.which("conda") else ("mamba" if shutil.which("mamba") else None)
    if not runner:
        return 1, "", "no conda/mamba"
    cmd = [runner, "install", "-y", "-p", env_path, "--dry-run"] + list(packages)
    return run_cmd_capture(cmd)


def clean_index_cache(*, yes):
    runner = "conda" if shutil.which("conda") else ("mamba" if shutil.which("mamba") else None)
    if not runner:
        return False
    cmd = [runner, "clean", "--index-cache"]
    if yes:
        cmd.append("-y")
    return run_cmd_live(cmd) == 0


def conda_info_json(*, show_json_output):
    runner = "conda" if shutil.which("conda") else ("mamba" if shutil.which("mamba") else None)
    if not runner:
        return None
    return run_json_cmd([runner, "info", "--json"], show_json_output=show_json_output)


def extract_pkgs_dirs(info):
    if not isinstance(info, dict):
        return []
    pkgs_dirs = info.get("pkgs_dirs")
    if isinstance(pkgs_dirs, list):
        return [p for p in pkgs_dirs if isinstance(p, str)]
    return []


def clean_cache_level(level):
    """
    Returns list of conda/mamba clean args for a given level.
    """
    if level == "safe":
        return [["--index-cache"], ["--tempfiles"]]
    if level == "targeted":
        return [["--tarballs"]]
    if level == "aggressive":
        return [["--all"]]
    return []


def conda_clean(args, *, yes):
    runner = "conda" if shutil.which("conda") else ("mamba" if shutil.which("mamba") else None)
    if not runner:
        return False
    cmd = [runner, "clean"] + list(args)
    if yes:
        cmd.append("-y")
    return run_cmd_live(cmd) == 0



import sys
import time


class Progress:
    def __init__(self, total, label, stream=None):
        self.total = total
        self.label = label
        self.stream = stream or sys.stderr
        self.start = time.time()
        self.last_len = 0

    def update(self, current):
        if self.total <= 0:
            return
        elapsed = time.time() - self.start
        rate = current / elapsed if elapsed > 0 else 0
        remaining = (self.total - current) / rate if rate > 0 else 0
        pct = int((current / self.total) * 100)
        msg = f"{self.label} {current}/{self.total} ({pct}%) ETA {remaining:0.1f}s"
        pad = " " * max(0, self.last_len - len(msg))
        self.stream.write("\r" + msg + pad)
        self.stream.flush()
        self.last_len = len(msg)

    def finish(self):
        if self.last_len:
            self.stream.write("\n")
            self.stream.flush()



from .conflicts import find_same_version_case_conflicts
from .search_parse import extract_search_results

__all__ = ["extract_search_results", "find_same_version_case_conflicts"]

__version__ = "0.2.0"



import json
import re
from pathlib import Path
from pathlib import PureWindowsPath


def extract_paths_from_text(text, *, env_prefix):
    """
    Best-effort extraction of conflicting paths from conda/mamba error output.
    We only return paths that appear to be inside the given env prefix.
    """
    if not text:
        return []

    def _looks_like_windows_abs(p):
        # Accept a normal Windows path (C:\...) and also log-escaped variants (C:\\...),
        # since conda/mamba errors often include doubled backslashes.
        return bool(re.match(r"^[A-Za-z]:\\", p or ""))

    def _normalize_windows_abs(p):
        # Logs often contain escaped backslashes (C:\\Anaconda3\\...) – unescape them.
        p = (p or "").replace("\\\\", "\\")
        # Normalize separators and case-insensitive comparison via a forward-slash form.
        return str(PureWindowsPath(p).as_posix()).lower()

    prefix_is_win = _looks_like_windows_abs(env_prefix)
    prefix_cmp = _normalize_windows_abs(env_prefix) if prefix_is_win else str(Path(env_prefix).resolve())
    if prefix_is_win:
        # Ensure consistent prefix matching (avoid C:/A matching C:/AB).
        prefix_cmp = prefix_cmp.rstrip("/")

    # Common patterns include quoted paths, or lines containing "path:".
    candidates = set()

    # Quoted windows paths: 'C:\\...'
    for m in re.finditer(r"['\"]([A-Za-z]:\\[^'\"]+)['\"]", text):
        candidates.add(m.group(1))
    # Unquoted windows paths: C:\...\something
    for m in re.finditer(r"([A-Za-z]:\\[^\\s\\r\\n]+)", text):
        candidates.add(m.group(1))
    # POSIX paths (for completeness)
    for m in re.finditer(r"(/[^\\s\\r\\n]+)", text):
        candidates.add(m.group(1))

    inside = []
    for p in candidates:
        if prefix_is_win and _looks_like_windows_abs(p):
            cand_norm = _normalize_windows_abs(p)
            if cand_norm == prefix_cmp or cand_norm.startswith(prefix_cmp + "/"):
                inside.append(str(PureWindowsPath(p.replace("\\\\", "\\"))))
            continue

        try:
            rp = str(Path(p).resolve())
        except Exception:
            continue
        # Normalize for case-insensitive comparison on Windows prefixes that are not Windows-abs.
        if rp.lower().startswith(str(prefix_cmp).lower()):
            inside.append(rp)
    return sorted(set(inside))


def build_conda_file_owner_map(env_prefix):
    """
    Map relative file paths (as stored in conda-meta JSON 'files') to package records.
    Returns dict[relpath] = {"name": ..., "version": ..., "build": ..., "record": ...}
    """
    owners = {}
    root = Path(env_prefix) / "conda-meta"
    if not root.exists():
        return owners

    for p in root.glob("*.json"):
        try:
            data = json.loads(p.read_text(encoding="utf-8", errors="ignore"))
        except Exception:
            continue
        if not isinstance(data, dict):
            continue
        files = data.get("files")
        if not isinstance(files, list):
            continue
        name = data.get("name")
        version = data.get("version")
        build = data.get("build") or data.get("build_string")
        for f in files:
            if not isinstance(f, str) or not f:
                continue
            rel = f.replace("\\", "/").lstrip("/")
            owners.setdefault(
                rel,
                {
                    "name": name,
                    "version": version,
                    "build": build,
                    "record": p.name,
                },
            )
    return owners


def to_relpath(env_prefix, abs_path):
    try:
        rel = str(Path(abs_path).resolve().relative_to(Path(env_prefix).resolve()))
    except Exception:
        return None
    return rel.replace("\\", "/")




import unittest

import env_repair as ed


class TestCaseConflicts(unittest.TestCase):
    def test_same_version_case_conflict_detected(self):
        entries = [
            {"name": "Automat", "version": "25.4.16", "channel": "pypi"},
            {"name": "automat", "version": "25.4.16", "channel": "conda-forge"},
            {"name": "Bottleneck", "version": "1.6.0", "channel": "pypi"},
            {"name": "bottleneck", "version": "1.6.0", "channel": "conda-forge"},
        ]
        pip_items, conda_force = ed.find_same_version_case_conflicts(entries)
        self.assertEqual({i["name"] for i in pip_items}, {"Automat", "Bottleneck"})
        self.assertEqual(set(conda_force), {"automat", "bottleneck"})

    def test_different_versions_not_selected(self):
        entries = [
            {"name": "Automat", "version": "25.4.16", "channel": "pypi"},
            {"name": "automat", "version": "25.4.15", "channel": "conda-forge"},
        ]
        pip_items, conda_force = ed.find_same_version_case_conflicts(entries)
        self.assertEqual(pip_items, [])
        self.assertEqual(conda_force, [])


if __name__ == "__main__":
    unittest.main()




import json
import tempfile
import unittest
from pathlib import Path

from env_repair.repair import _conda_meta_owns_distinfo


class TestCaseConflictCondaOwned(unittest.TestCase):
    def test_conda_meta_owns_distinfo_true(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            meta = env / "conda-meta"
            meta.mkdir(parents=True, exist_ok=True)
            record = {
                "name": "pydrive",
                "version": "1.3.1",
                "build": "py_1",
                "files": [
                    "Lib/site-packages/PyDrive-1.3.1.dist-info/METADATA",
                    "Lib/site-packages/pydrive/__init__.py",
                ],
            }
            (meta / "pydrive-1.3.1-py_1.json").write_text(json.dumps(record), encoding="utf-8")
            self.assertTrue(
                _conda_meta_owns_distinfo(
                    str(env),
                    conda_pkg_name="pydrive",
                    dist_name="PyDrive",
                    version="1.3.1",
                )
            )

    def test_conda_meta_owns_distinfo_false(self):
        with tempfile.TemporaryDirectory() as td:
            env = Path(td)
            meta = env / "conda-meta"
            meta.mkdir(parents=True, exist_ok=True)
            record = {"name": "pydrive", "version": "1.3.1", "build": "py_1", "files": []}
            (meta / "pydrive-1.3.1-py_1.json").write_text(json.dumps(record), encoding="utf-8")
            self.assertFalse(
                _conda_meta_owns_distinfo(
                    str(env),
                    conda_pkg_name="pydrive",
                    dist_name="PyDrive",
                    version="1.3.1",
                )
            )


if __name__ == "__main__":
    unittest.main()




import json
import tempfile
import unittest
from pathlib import Path

from env_repair.repair import _blacklist_add, _extract_incompatible_specs, _load_adopt_pip_blacklist, _save_adopt_pip_blacklist


class TestAdoptPipBlacklist(unittest.TestCase):
    def test_extracts_incompatible_specs(self):
        text = """
error    libmamba Could not solve for environment specs
    The following packages are incompatible
    ├─ langchain-chroma =* * is installable with the potential options
    └─ pin on python =3.13 * is not installable because it requires
"""
        names = _extract_incompatible_specs(text)
        self.assertIn("langchain-chroma", names)
        self.assertIn("pin", names)

    def test_blacklist_roundtrip(self):
        with tempfile.TemporaryDirectory() as td:
            cwd = Path(td)
            (cwd / ".env_repair").mkdir(parents=True, exist_ok=True)
            prev = Path.cwd()
            try:
                # Change CWD so the helper path writes into temp folder.
                import os

                os.chdir(cwd)
                blocked = {}
                self.assertTrue(
                    _blacklist_add(
                        blocked,
                        pip_name="langchain-chroma",
                        pip_version="9.9.9",
                        conda_name="langchain-chroma",
                        reason="solver_incompatible",
                    )
                )
                _save_adopt_pip_blacklist(blocked)
                loaded = _load_adopt_pip_blacklist()
                self.assertIn("langchain-chroma", loaded)
                self.assertIn("9.9.9", loaded["langchain-chroma"])
            finally:
                os.chdir(prev)


if __name__ == "__main__":
    unittest.main()



import concurrent.futures
import os
import json
import subprocess
from pathlib import Path

from .conda_ops import conda_install, conda_install_capture, conda_remove, get_env_package_entries, is_conda_env
from .conda_config import load_conda_channels
from .discovery import discover_envs, get_python_exe, select_envs, which
from .naming import normalize_name
from .pip_ops import pip_get_version, pip_reinstall, pip_uninstall
from .progress import Progress
from .subprocess_utils import run_json_cmd

CRITICAL_PACKAGES = {
    "pip",
    "setuptools",
    "wheel",
    "numpy",
    "scipy",
    "pandas",
    "matplotlib",
    "requests",
    "certifi",
    "ssl",
}

CRITICAL_PIP_PACKAGES = {"pip", "setuptools", "wheel"}


def _verify_imports_blacklist_path():
    return Path(".env_repair") / "verify_imports_blacklist.json"


def _load_verify_imports_blacklist():
    path = _verify_imports_blacklist_path()
    if not path.exists():
        return {}
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return {}
    if not isinstance(data, dict):
        return {}
    blocked = data.get("blocked")
    if not isinstance(blocked, dict):
        return {}
    out = {}
    for pyver, names in blocked.items():
        if not isinstance(pyver, str) or not isinstance(names, dict):
            continue
        out.setdefault(pyver, {})
        for name, meta in names.items():
            if not isinstance(name, str):
                continue
            out[pyver][normalize_name(name)] = meta if isinstance(meta, dict) else {}
    return out


def _save_verify_imports_blacklist(blocked):
    path = _verify_imports_blacklist_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    payload = {"blocked": blocked}
    path.write_text(json.dumps(payload, indent=2, sort_keys=True), encoding="utf-8")


def _blacklist_add(blocked, *, pyver, conda_name, reason):
    if not pyver or not conda_name:
        return False
    blocked.setdefault(pyver, {})
    blocked[pyver].setdefault(
        normalize_name(conda_name),
        {
            "conda": conda_name,
            "reason": reason,
        },
    )
    return True


def _read_metadata_name(dist_info):
    meta = dist_info / "METADATA"
    if not meta.exists():
        return None
    try:
        for line in meta.read_text(encoding="utf-8", errors="ignore").splitlines():
            if not line:
                break
            if line.lower().startswith("name:"):
                val = line.split(":", 1)[1].strip()
                return val or None
    except OSError:
        return None
    return None


def _installed_by(dist_info):
    installer_path = dist_info / "INSTALLER"
    if installer_path.exists():
        content = installer_path.read_text(encoding="utf-8", errors="ignore").strip().lower()
        if "conda" in content:
            return "conda"
        if "pip" in content:
            return "pip"
    if (dist_info / "direct_url.json").exists():
        return "pip"
    return "unknown"


def _classify_dist(dist_info, *, conda_entries_by_name):
    dist_name = _read_metadata_name(dist_info) or dist_info.name.split("-")[0]
    key = normalize_name(dist_name)
    entry = conda_entries_by_name.get(key)
    if entry:
        channel = (entry.get("channel") or "").lower()
        if "pypi" in channel:
            return "pip", entry.get("name") or dist_name
        return "conda", entry.get("name") or dist_name
    return _installed_by(dist_info), dist_name


def _extract_replacement_from_search_json(search_json, *, target):
    """
    Extract the "replacement" package name from a conda/mamba search --json payload.

    We interpret a "replacement" as: a dependency list containing the target package.
    """
    if not isinstance(search_json, dict):
        return None
    result = search_json.get("result") or {}
    pkgs = result.get("pkgs") or []
    if not isinstance(pkgs, list):
        return None
    for pkg in pkgs:
        if not isinstance(pkg, dict):
            continue
        deps = pkg.get("depends") or []
        if not isinstance(deps, list):
            continue
        for dep in deps:
            if not isinstance(dep, str):
                continue
            name = dep.split(" ", 1)[0].strip()
            if name == target:
                return target
    return None


def _maybe_replace_deprecated_wrapper(conda_name, *, manager, debug):
    """
    Handle known deprecated wrapper packages by installing their new backend instead.

    Currently supported:
      - conda-token -> anaconda-auth (conda-token >=0.7.0 depends on it)
    """
    key = normalize_name(conda_name)
    if key != "conda-token":
        return None

    if manager not in ("conda", "mamba"):
        return None

    # Search in defaults first because conda-token/anaconda-auth are hosted on pkgs/main.
    data = run_json_cmd([manager, "search", "-c", "defaults", "conda-token", "--json"], show_json_output=debug)
    if not data:
        return None
    if _extract_replacement_from_search_json(data, target="anaconda-auth"):
        return {"install": "anaconda-auth", "remove": "conda-token"}
    return None


def _collect_wrapper_removals(plan, *, conda_installed):
    """
    Determine which planned uninstall items we can execute after successful installs.

    `conda_installed` is a set of conda package names that were successfully installed/reinstalled in this run.
    """
    conda_remove_pkgs = []
    pip_remove_pkgs = []
    installed_norm = {normalize_name(p) for p in (conda_installed or set())}

    for item in plan or []:
        uninstall = item.get("uninstall") or {}
        remove_name = uninstall.get("name")
        remove_kind = uninstall.get("kind")
        required = item.get("name")  # the replacement we attempted to install
        if not remove_name or not remove_kind or not required:
            continue
        if normalize_name(required) not in installed_norm:
            continue
        if remove_kind == "conda":
            conda_remove_pkgs.append(remove_name)
        elif remove_kind == "pip":
            pip_remove_pkgs.append(remove_name)

    return sorted(set(conda_remove_pkgs)), sorted(set(pip_remove_pkgs))


def _collect_explicit_removals(plan):
    """
    Collect removal-only actions requested by the fix planner.

    These do not depend on installing a replacement in the same run; they are executed directly.
    """
    conda_rm = []
    pip_rm = []
    for item in plan or []:
        if item.get("kind") != "remove":
            continue
        uninstall = item.get("uninstall") or {}
        name = uninstall.get("name")
        kind = uninstall.get("kind")
        if not name or not kind:
            continue
        if kind == "conda":
            conda_rm.append(name)
        elif kind == "pip":
            pip_rm.append(name)
    return sorted(set(conda_rm)), sorted(set(pip_rm))


def _is_removed_collections_mapping_error(error):
    if not error:
        return False
    e = str(error)
    return ("cannot import name 'Mapping' from 'collections'" in e) or ('cannot import name "Mapping" from "collections"' in e)


def _is_boto_vendored_six_moves_error(error):
    if not error:
        return False
    return "No module named 'boto.vendored.six.moves'" in str(error)


def _extract_missing_module_name(error):
    if not error:
        return None
    e = str(error)
    # Common forms:
    #   ModuleNotFoundError: No module named 'numpy'
    #   ModuleNotFoundError: No module named numpy
    marker = "No module named"
    if marker not in e:
        return None
    try:
        tail = e.split(marker, 1)[1].strip()
    except Exception:
        return None
    if not tail:
        return None
    # Strip leading ':' or quotes
    if tail.startswith(":"):
        tail = tail[1:].strip()
    if tail.startswith("'") or tail.startswith('"'):
        q = tail[0]
        end = tail.find(q, 1)
        if end > 1:
            tail = tail[1:end]
    # Only take the first token/module path component (numpy.linalg -> numpy)
    mod = tail.split()[0].strip().strip("'\"")
    if not mod:
        return None
    return mod.split(".", 1)[0]


def _python_major_minor(python_exe):
    try:
        proc = subprocess.run(
            [python_exe, "-c", "import sys;print(f'{sys.version_info[0]}.{sys.version_info[1]}')"],
            capture_output=True,
            text=True,
            timeout=15,
            check=False,
        )
    except Exception:
        return None
    if proc.returncode != 0:
        return None
    val = (proc.stdout or "").strip()
    return val or None


def _extract_solver_offenders(output_text):
    """
    Best-effort parse of libmamba solver output to extract offending specs.
    """
    import re

    if not output_text:
        return set()
    offenders = set()
    # Lines like: "└─ tables =* * does not exist"
    for m in re.finditer(r"(?:^|\n)\s*[├└]─\s*([A-Za-z0-9_.-]+)\s", output_text):
        offenders.add(m.group(1))
    return offenders


def _should_skip_failure(f):
    err = (f.get("error") or "").lower()
    name = (f.get("import") or "").lower()
    if os.name == "nt":
        if name == "sh" and ("not supported on windows" in err or "windows" in err):
            return True
        if "only supported on linux" in err or "only supported on macos" in err:
            return True
        if "no module named 'fcntl'" in err or "no module named fcntl" in err:
            return True
        if name == "ptyprocess" and ("no module named 'pty'" in err or "no module named pty" in err):
            return True
    return False


def _is_idna_ssl_match_hostname_error(err):
    if not err:
        return False
    e = err.lower()
    return "idna_ssl" in e and "ssl" in e and "has no attribute" in e and "match_hostname" in e


def _is_jedi_common_shadowing_error(err):
    if not err:
        return False
    e = err.lower()
    return (
        "cannot import name" in e
        and "from 'jedi.common'" in e
        and ("jedi\\common\\__init__.py" in e or "jedi/common/__init__.py" in e)
    )


def _conda_pkg_has_site_packages_files(prefix, conda_name):
    """
    Return True if conda-meta for `conda_name` lists any Lib/site-packages files.
    Useful to distinguish non-Python conda packages that can't fix a Python import.
    """
    try:
        meta_dir = Path(prefix) / "conda-meta"
        pats = sorted(meta_dir.glob(f"{conda_name}-*.json"))
        if not pats:
            return False
        meta_path = pats[-1]
        data = json.loads(meta_path.read_text(encoding="utf-8", errors="ignore"))
        files = data.get("files") or []
        for f in files:
            if isinstance(f, str) and "Lib/site-packages" in f.replace("\\", "/"):
                return True
    except Exception:
        return False
    return False


def _cleanup_jedi_common_pkg_dir(python_exe):
    """
    If both `jedi/common.py` and `jedi/common/` exist, the package directory can shadow the module.
    In broken environments this can make `from jedi.common import indent_block/monkeypatch` fail.
    Remove the `jedi/common/` directory when `jedi/common.py` is present.
    """
    try:
        import shutil
        from .discovery import get_site_packages

        for sp in get_site_packages(python_exe) or []:
            sp = Path(sp)
            module_path = sp / "jedi" / "common.py"
            pkg_dir = sp / "jedi" / "common"
            if module_path.exists() and pkg_dir.exists() and pkg_dir.is_dir():
                shutil.rmtree(pkg_dir, ignore_errors=True)
                return True
        return True
    except Exception:
        return False


def attempt_fix(failures, python_exe, env_path, manager, *, base_prefix, debug):
    if not failures:
        return {"ok": True, "actions": []}

    is_conda = is_conda_env(env_path)
    entries = get_env_package_entries(env_path, manager, show_json_output=debug) if (is_conda and manager) else []
    conda_entries_by_name = {normalize_name(e.get("name") or ""): e for e in entries if isinstance(e, dict)}
    initially_installed = set(conda_entries_by_name.keys())
    pyver = _python_major_minor(python_exe) or "unknown"
    blacklist = _load_verify_imports_blacklist()
    blocked_for_py = blacklist.get(pyver) or {}
    blocked_names = set(blocked_for_py.keys())
    configured_channels = load_conda_channels(base_prefix=base_prefix, has_conda=which("conda") is not None, show_json_output=debug)
    nodefaults = "nodefaults" in configured_channels
    has_defaults = ("defaults" in configured_channels) or ("anaconda" in configured_channels)

    # Stage 0: If failures are caused by a missing dependency module, fix that first.
    missing = {}
    for f in failures:
        mod = _extract_missing_module_name(f.get("error"))
        if mod:
            missing[normalize_name(mod)] = missing.get(normalize_name(mod), 0) + 1
    if missing:
        conda_targets = []
        pip_targets = []
        for mod_norm in sorted(missing.keys()):
            entry = conda_entries_by_name.get(mod_norm) or {}
            if entry:
                conda_name = entry.get("name") or mod_norm
                # Some conda packages (e.g. flatbuffers) are non-Python libs and won't fix a missing import.
                if manager and not _conda_pkg_has_site_packages_files(env_path, conda_name):
                    pip_targets.append(mod_norm)
                else:
                    conda_targets.append(conda_name)
                continue
            # If listed as a pip entry via conda list (channel pypi), treat as pip.
            if mod_norm in initially_installed and ((entry.get("channel") or "").lower() == "pypi"):
                pip_targets.append(entry.get("name") or mod_norm)
        if conda_targets or pip_targets:
            print("\nPriority: missing modules detected; repairing dependencies first...")
            if conda_targets and manager:
                conda_targets = sorted(set(conda_targets))
                print("Conda dependency targets:", ", ".join(conda_targets))
                ok_dep = conda_install(
                    env_path,
                    conda_targets,
                    manager,
                    configured_channels,
                    ignore_pinned=False,
                    force_reinstall=True,
                )
                if not ok_dep:
                    actions = [{"action": "conda_dependency_reinstall", "ok": False, "packages": conda_targets}]
                    return {"ok": False, "plan": [], "actions": actions}
            if pip_targets:
                pip_targets = sorted(set(pip_targets))
                print("Pip dependency targets:", ", ".join(pip_targets))
                for pkg in pip_targets:
                    pip_reinstall(
                        python_exe,
                        pkg,
                        no_deps=bool(is_conda),
                        only_binary=os.name == "nt",
                        ignore_installed=bool(is_conda),
                    )

            # Recheck failures after dependency repair and keep only remaining ones.
            still = []
            for f in failures:
                imp = f.get("import")
                if not imp or not isinstance(imp, str):
                    still.append(f)
                    continue
                ok, err = check_import(imp, python_exe)
                if not ok:
                    f["error"] = err
                    still.append(f)
            failures = still
            if not failures:
                return {"ok": True, "plan": [], "actions": [{"action": "dependency_repair_only", "ok": True}]}

    by_dist = {}
    for f in failures:
        dist_info = f.get("dist_path")
        if isinstance(dist_info, Path):
            by_dist.setdefault(dist_info, f)

    plan = []
    actions = []
    for dist_info, f in sorted(by_dist.items(), key=lambda kv: str(kv[0]).lower()):
        failed_import = f.get("import")
        if _should_skip_failure(f):
            plan.append({"dist": dist_info.name, "kind": "skip", "name": None, "import": failed_import, "reason": "platform/expected"})
            continue
        kind, name = _classify_dist(dist_info, conda_entries_by_name=conda_entries_by_name)

        # The verify-imports blacklist is about conda solver failures. Do not block pip repairs.
        if kind == "conda" and name and normalize_name(name) in blocked_names:
            # If an import is crashing the interpreter, prefer removing the package rather than looping forever.
            err_lower = (f.get("error") or "").lower()
            if "fatal python error" in err_lower:
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "remove",
                        "name": None,
                        "import": failed_import,
                        "reason": f"blacklisted for python {pyver} + fatal import crash; removing package",
                        "uninstall": {"kind": "conda", "name": name},
                    }
                )
                continue
            plan.append(
                {
                    "dist": dist_info.name,
                    "kind": "skip",
                    "name": None,
                    "import": failed_import,
                    "reason": f"blacklisted for python {pyver} (previous solver failure)",
                }
            )
            continue

        # idna_ssl is a legacy compatibility package; on modern Python it can break due to removed APIs.
        if name and normalize_name(name) in ("idna-ssl", "idna_ssl") and _is_idna_ssl_match_hostname_error(f.get("error")):
            uninstall_kind = "conda" if kind == "conda" else "pip"
            plan.append(
                {
                    "dist": dist_info.name,
                    "kind": "remove",
                    "name": None,
                    "import": failed_import,
                    "reason": "obsolete/incompatible: idna_ssl relies on ssl.match_hostname (removed); removing",
                    "uninstall": {"kind": uninstall_kind, "name": name},
                }
            )
            continue

        # Jedi should be importable on Windows; if `jedi/common/` shadows `jedi/common.py`, clean it up.
        if name and normalize_name(name) == "jedi" and _is_jedi_common_shadowing_error(f.get("error")):
            plan.append(
                {
                    "dist": dist_info.name,
                    "kind": "cleanup",
                    "name": "jedi-common-dir",
                    "import": failed_import,
                    "reason": "jedi.common shadowing: remove jedi/common/ so jedi/common.py wins",
                }
            )
            continue

        # If a conda package cannot possibly fix a missing Python module (no site-packages files),
        # fall back to pip for the import name.
        missing_mod = _extract_missing_module_name(f.get("error"))
        if kind == "conda" and name and missing_mod and normalize_name(missing_mod) == normalize_name(failed_import or ""):
            if not _conda_pkg_has_site_packages_files(env_path, name):
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "pip",
                        "name": missing_mod,
                        "import": failed_import,
                        "reason": f"conda '{name}' has no site-packages files; installing '{missing_mod}' via pip",
                    }
                )
                continue

        # attrdict is archived/inactive and breaks on modern Python because collections.Mapping was removed.
        # Replace it with attrdict3 (drop-in fork) when we see this signature.
        if name and normalize_name(name) == "attrdict" and _is_removed_collections_mapping_error(f.get("error")):
            plan.append(
                {
                    "dist": dist_info.name,
                    "kind": "conda",
                    "name": "attrdict3",
                    "import": failed_import,
                    "reason": "deprecated/broken: replace attrdict with attrdict3 (collections.Mapping removed)",
                    "uninstall": {"kind": "conda" if kind == "conda" else "pip", "name": "attrdict"},
                }
            )
            continue

        # Handle deprecated wrappers even if defaults are disabled:
        # - if the replacement is already installed, just remove the wrapper
        # - otherwise try to install the replacement and then remove the wrapper
        if name and normalize_name(name) == "conda-token":
            uninstall_kind = "conda" if kind == "conda" else "pip"
            if normalize_name("anaconda-auth") in initially_installed:
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "remove",
                        "name": None,
                        "import": failed_import,
                        "reason": "deprecated-wrapper: conda-token already replaced; remove conda-token (no reinstall)",
                        "uninstall": {"kind": uninstall_kind, "name": "conda-token"},
                    }
                )
                continue
            replacement = _maybe_replace_deprecated_wrapper(name, manager=manager, debug=debug)
            if replacement:
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "conda",
                        "name": replacement["install"],
                        "import": failed_import,
                        "reason": f"deprecated-wrapper: {name} -> {replacement['install']}",
                        "uninstall": {"kind": uninstall_kind, "name": replacement["remove"]},
                    }
                )
                continue

        # boto2 can be effectively unfixable on modern Python; this specific error indicates its vendored six can't provide moves.
        # boto3 is the successor (but does not provide the `boto` module). If boto3 is present, remove boto2 to stop the loop.
        if name and normalize_name(name) == "boto" and _is_boto_vendored_six_moves_error(f.get("error")):
            if normalize_name("boto3") in initially_installed:
                uninstall_kind = "conda" if kind == "conda" else "pip"
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "remove",
                        "name": None,
                        "import": failed_import,
                        "reason": "unfixable: boto (boto2) fails (vendored six.moves); remove boto (boto3 already present)",
                        "uninstall": {"kind": uninstall_kind, "name": "boto"},
                    }
                )
                continue
            plan.append(
                {
                    "dist": dist_info.name,
                    "kind": "skip",
                    "name": None,
                    "import": failed_import,
                    "reason": "unfixable: boto (boto2) fails (vendored six.moves); migrate to boto3 or pin older Python",
                }
            )
            continue

        if kind == "conda" and nodefaults and not has_defaults:
            entry = conda_entries_by_name.get(normalize_name(name or "")) or {}
            ch = (entry.get("channel") or "").lower()
            if ch.startswith("pkgs"):
                # Even if defaults are disabled, we can still try to reinstall from the configured channels
                # (typically conda-forge). If that fails, the solver-offender logic will blacklist it.
                plan.append(
                    {
                        "dist": dist_info.name,
                        "kind": "conda",
                        "name": name,
                        "import": failed_import,
                        "reason": "defaults-disabled: attempting reinstall from configured channels",
                    }
                )
                continue
        plan.append({"dist": dist_info.name, "kind": kind, "name": name, "import": failed_import, "reason": None})

    def _print_fix_plan(plan_items):
        items = list(plan_items or [])
        if not items:
            return
        print("\nFix plan:")
        for item in items:
            kind = item.get("kind") or "?"
            name = item.get("name")
            dist = item.get("dist") or "?"
            reason = item.get("reason")
            uninstall = item.get("uninstall") or {}
            remove_name = uninstall.get("name")
            remove_kind = uninstall.get("kind")
            if kind == "remove" and remove_name and remove_kind:
                line = f"  - {dist}: remove via {remove_kind}: {remove_name}"
                if reason:
                    line += f" [{reason}]"
            else:
                line = f"  - {dist}: {kind}"
                if name:
                    line += f" -> {name}"
                if reason:
                    line += f" [{reason}]"
                if remove_name and remove_kind:
                    line += f" (then remove via {remove_kind}: {remove_name})"
            print(line)

    _print_fix_plan(plan)

    ok_all = True

    # Explicit removals (no install dependency).
    explicit_conda_rm, explicit_pip_rm = _collect_explicit_removals(plan)
    if explicit_conda_rm:
        if is_conda and manager:
            refreshed = get_env_package_entries(env_path, manager, show_json_output=debug)
            present = {normalize_name((e or {}).get("name") or "") for e in refreshed if isinstance(e, dict)}
            conda_rm_present = [p for p in explicit_conda_rm if normalize_name(p) in present]
            if conda_rm_present:
                print(f"\nRemoving {len(conda_rm_present)} package(s) via {manager} (remove-only)...")
                print("Conda remove targets:", ", ".join(conda_rm_present))
                ok_rm = conda_remove(env_path, conda_rm_present, manager)
                ok_all = ok_all and ok_rm
                actions.append({"action": "conda_remove", "count": len(conda_rm_present), "ok": ok_rm, "packages": conda_rm_present})
        else:
            ok_all = False
            actions.append({"action": "error", "reason": "remove-only requested but no conda/mamba manager available"})
    if explicit_pip_rm and python_exe:
        print(f"\nRemoving {len(explicit_pip_rm)} package(s) via pip (remove-only)...")
        print("Pip remove targets:", ", ".join(explicit_pip_rm))
        ok_rm = pip_uninstall(python_exe, explicit_pip_rm)
        ok_all = ok_all and ok_rm
        actions.append({"action": "pip_uninstall", "count": len(explicit_pip_rm), "ok": ok_rm, "packages": explicit_pip_rm})

    cleanup_items = [p for p in plan if p.get("kind") == "cleanup" and p.get("name")]
    for item in cleanup_items:
        if normalize_name(item.get("name") or "") != "jedi-common-dir":
            continue
        print("\nCleaning up jedi/common shadowing (site-packages)...")
        ok_clean = _cleanup_jedi_common_pkg_dir(python_exe)
        ok_all = ok_all and ok_clean
        actions.append({"action": "cleanup_jedi_common_dir", "ok": ok_clean})

    def _install_conda_batched(packages):
        """
        Try to install all packages in one go for speed.

        If the solver fails, bisect into batches (never single-package installs),
        so we can still make progress without going fully one-by-one.
        """
        pkgs = sorted({p for p in packages if p})
        if not pkgs:
            return True, [], []

        def _try(group):
            ok_group, out, err = conda_install_capture(
                env_path,
                group,
                manager,
                configured_channels,
                ignore_pinned=False,
                force_reinstall=True,
            )
            if ok_group:
                return group, []

            # If libmamba explicitly names offender specs, blacklist and retry without them.
            offenders = sorted({p for p in _extract_solver_offenders((out or "") + "\n" + (err or "")) if p in set(group)})
            if offenders:
                for off in offenders:
                    _blacklist_add(blacklist, pyver=pyver, conda_name=off, reason="solver_incompatible")
                _save_verify_imports_blacklist(blacklist)
                remaining = [p for p in group if p not in set(offenders)]
                if len(remaining) >= 2:
                    print(f"Solver failed; skipping {len(offenders)} incompatible package(s): {', '.join(offenders)}")
                    ok2, _out2, _err2 = conda_install_capture(
                        env_path,
                        remaining,
                        manager,
                        configured_channels,
                        ignore_pinned=False,
                        force_reinstall=True,
                    )
                    if ok2:
                        return remaining, offenders
                # If we can't retry meaningfully, keep original failure.
            if len(group) <= 2:
                # Never split into single-package installs; if 1-2 fail, mark as failed.
                return [], group
            mid = len(group) // 2
            left = group[:mid]
            right = group[mid:]
            # If a split would create a singleton, don't go further.
            if len(left) < 2 or len(right) < 2:
                return [], group
            ok_left, bad_left = _try(left)
            ok_right, bad_right = _try(right)
            return ok_left + ok_right, bad_left + bad_right

        installed, failed = _try(pkgs)
        installed = sorted(set(installed))
        failed = sorted(set(failed))
        return len(failed) == 0, installed, failed

    def _relink_after_conda_remove(plan_items, *, installed_set):
        """
        After `conda remove`, re-force-reinstall any replacement packages that may share files
        with the removed package(s). This mitigates conda's lack of per-file ownership tracking.
        """
        if not manager:
            return True
        to_relink = []
        installed_norm = {normalize_name(p) for p in (installed_set or set())}
        for item in plan_items or []:
            uninstall = item.get("uninstall") or {}
            if uninstall.get("kind") != "conda":
                continue
            replacement = item.get("name")
            if not replacement:
                continue
            if normalize_name(replacement) in installed_norm:
                to_relink.append(replacement)
        pkgs = sorted(set(to_relink))
        if not pkgs:
            return True
        print(f"\nRelinking {len(pkgs)} replacement package(s) via {manager} (--force-reinstall)...")
        print("Conda relink targets:", ", ".join(pkgs))
        return conda_install(env_path, pkgs, manager, configured_channels, ignore_pinned=False, force_reinstall=True)

    conda_items = [p for p in plan if p["kind"] == "conda" and p["name"]]
    if conda_items and manager:
        pkgs = [p["name"] for p in conda_items]
        print(f"\nReinstalling {len(pkgs)} packages via {manager} (batched)...")
        print("Conda targets:", ", ".join(sorted(set(pkgs))))
        ok, installed, failed = _install_conda_batched(pkgs)
        ok_all = ok_all and ok
        actions.append({"action": "conda_reinstall_batch", "count": len(pkgs), "ok": ok, "installed": installed, "failed": failed})

        # If the replacement was installed successfully, remove deprecated wrappers.
        conda_rm, pip_rm = _collect_wrapper_removals(plan, conda_installed=set(installed) | initially_installed)
        if conda_rm:
            refreshed = get_env_package_entries(env_path, manager, show_json_output=debug) if is_conda else []
            present = {normalize_name((e or {}).get("name") or "") for e in refreshed if isinstance(e, dict)}
            conda_rm_present = [p for p in conda_rm if normalize_name(p) in present]
            if conda_rm_present:
                print(f"\nRemoving {len(conda_rm_present)} deprecated wrapper package(s) via {manager}...")
                ok_rm = conda_remove(env_path, conda_rm_present, manager)
                ok_all = ok_all and ok_rm
                actions.append({"action": "conda_remove", "count": len(conda_rm_present), "ok": ok_rm, "packages": conda_rm_present})
                if ok_rm:
                    ok_relink = _relink_after_conda_remove(plan, installed_set=set(installed) | initially_installed)
                    ok_all = ok_all and ok_relink
                    actions.append({"action": "conda_relink", "ok": ok_relink})
        if pip_rm and python_exe:
            print(f"\nRemoving {len(pip_rm)} deprecated wrapper package(s) via pip...")
            ok_rm = pip_uninstall(python_exe, pip_rm)
            ok_all = ok_all and ok_rm
            actions.append({"action": "pip_uninstall", "count": len(pip_rm), "ok": ok_rm, "packages": pip_rm})

    # Wrapper removals that don't require installing anything (replacement already present).
    conda_rm, pip_rm = _collect_wrapper_removals(plan, conda_installed=initially_installed)
    if conda_rm and manager:
        refreshed = get_env_package_entries(env_path, manager, show_json_output=debug) if is_conda else []
        present = {normalize_name((e or {}).get("name") or "") for e in refreshed if isinstance(e, dict)}
        conda_rm_present = [p for p in conda_rm if normalize_name(p) in present]
        if conda_rm_present:
            print(f"\nRemoving {len(conda_rm_present)} deprecated wrapper package(s) via {manager}...")
            ok_rm = conda_remove(env_path, conda_rm_present, manager)
            ok_all = ok_all and ok_rm
            actions.append({"action": "conda_remove", "count": len(conda_rm_present), "ok": ok_rm, "packages": conda_rm_present})
            if ok_rm:
                ok_relink = _relink_after_conda_remove(plan, installed_set=initially_installed)
                ok_all = ok_all and ok_relink
                actions.append({"action": "conda_relink", "ok": ok_relink})
    if pip_rm and python_exe:
        print(f"\nRemoving {len(pip_rm)} deprecated wrapper package(s) via pip...")
        ok_rm = pip_uninstall(python_exe, pip_rm)
        ok_all = ok_all and ok_rm
        actions.append({"action": "pip_uninstall", "count": len(pip_rm), "ok": ok_rm, "packages": pip_rm})

    pip_items = [p for p in plan if p["kind"] in ("pip", "unknown") and p["name"]]
    if pip_items:
        no_deps = bool(is_conda)
        why = " (no-deps in conda env)" if no_deps else ""
        only_binary = os.name == "nt"
        ignore_installed = bool(is_conda)
        extra = (why + (", wheels-only" if only_binary else "")) if (why or only_binary) else ""
        print(f"\nReinstalling {len(pip_items)} packages via pip{extra}...")
        print("Pip targets:", ", ".join(sorted(set(p['name'] for p in pip_items if p.get('name')))))
        for item in pip_items:
            pkg = item["name"]
            imp = item.get("import")
            print(f"  - {pkg}")
            ok = pip_reinstall(
                python_exe,
                pkg,
                no_deps=no_deps,
                only_binary=only_binary,
                ignore_installed=ignore_installed,
            )
            ok_all = ok_all and ok
            actions.append(
                {
                    "action": "pip_reinstall",
                    "package": pkg,
                    "ok": ok,
                    "no_deps": no_deps,
                    "only_binary": only_binary,
                    "ignore_installed": ignore_installed,
                }
            )

            # If the reinstall succeeded but the import still fails, remove the package (pip-managed only).
            if ok and imp and imp.isidentifier():
                ok_imp, err_imp = check_import(imp, python_exe)
                actions.append({"action": "pip_recheck_import", "package": pkg, "import": imp, "ok": ok_imp})
                if not ok_imp:
                    if normalize_name(pkg) in {normalize_name(p) for p in CRITICAL_PIP_PACKAGES}:
                        print(f"    ! Import still fails for {imp}, but refusing to uninstall critical pip package: {pkg}")
                        actions.append({"action": "skip_uninstall_critical_pip", "package": pkg, "import": imp})
                        ok_all = False
                        continue

                    ver = pip_get_version(python_exe, pkg)
                    ver_str = ver or "unknown"
                    print(f"    ! Import still fails after reinstall; removing via pip: {pkg}=={ver_str}")
                    if err_imp:
                        first_lines = err_imp.splitlines()[:8]
                        for line in first_lines:
                            print(f"      {line}")
                    ok_rm = pip_uninstall(python_exe, [pkg])
                    ok_all = ok_all and ok_rm
                    actions.append(
                        {
                            "action": "pip_uninstall_after_failed_reinstall",
                            "package": pkg,
                            "version": ver,
                            "import": imp,
                            "import_error": err_imp,
                            "ok": ok_rm,
                        }
                    )

    skipped = [p for p in plan if p["kind"] == "skip"]
    for item in skipped:
        actions.append({"action": "skip", "dist": item["dist"], "reason": item["reason"]})

    if is_conda and not manager and conda_items:
        ok_all = False
        actions.append({"action": "error", "reason": "conda env but no manager found"})

    return {"ok": ok_all, "plan": plan, "actions": actions}


def parse_record_file(record_path):
    """
    Parse a RECORD file to find the top-level importable packages.
    Heuristic: Look for files in top-level directories (excluding .dist-info/ etc).
    """
    try:
        content = record_path.read_text(encoding="utf-8", errors="ignore")
    except Exception:
        return set()

    top_levels = set()
    for line in content.splitlines():
        # RECORD format: path,sha256=...,size
        parts = line.split(",")
        if not parts:
            continue
        path = parts[0].strip()
        
        # Windows/Linux path normalization
        path = path.replace("\\", "/")
        
        # Filter noise
        if path.startswith("..") or "/site-packages/" in path:
            continue
        if ".dist-info/" in path or ".egg-info/" in path:
            continue
        
        # Start of path is usually the package name
        if "/" in path:
            top = path.split("/")[0]
            # Ignore __pycache__
            if top == "__pycache__":
                continue
            # If it looks like a valid identifier, add it
            if top.isidentifier():
                top_levels.add(top)
        else:
            # File in root (e.g. 'six.py')
            if path.endswith(".py"):
                top_levels.add(path[:-3])
                
    return sorted(top_levels)

def get_toplevel_imports(dist):
    """
    Extract the top-level import names from a distribution.
    Strategie:
    1. Parse RECORD (most accurate for file-based structure).
    2. Fallback to top_level.txt (can contain virtual modules or old metadata).
    3. Fallback to distribution name guessing.
    """
    # 1. Try RECORD (used by wheels, conda, etc.) - PREFERRED
    record = dist / "RECORD"
    if record.exists():
        found = parse_record_file(record)
        if found:
            return found

    # 2. Try top_level.txt (standard but sometimes lying, e.g. for CFFI bindings)
    top_level = dist / "top_level.txt"
    if top_level.exists():
        try:
            return [line.strip() for line in top_level.read_text().splitlines() if line.strip()]
        except Exception:
            pass
            
    # 3. Fallback: derive from distribution name
    name = dist.name.split("-")[0].replace("_", ".")
    return [name] if name else []

def check_import(package_name, python_exe):
    """
    Run `python -c "import <name>"` via subprocess.
    Returns (ok: bool, error_message: str|None).
    """
    cmd = [python_exe, "-c", f"import {package_name}"]
    try:
        proc = subprocess.run(cmd, capture_output=True, text=True, timeout=30)
        if proc.returncode == 0:
            return True, None
        return False, proc.stderr.strip()
    except subprocess.TimeoutExpired:
        return False, "Import timed out (>30s)"
    except Exception as e:
        return False, str(e)

def verify_imports(args):
    # `--debug` should show command lines, but avoid dumping huge `--json` payloads by default.
    show_json_output = False
    lang = "auto"
    
    all_envs, base_prefix, manager = discover_envs(show_json_output=show_json_output)
    targets = select_envs(all_envs, [args.env] if args.env else [], base_prefix)
    
    if not targets:
        return {"ok": False, "exit_code": 2, "error": "no target env"}
    
    env_path = targets[0]
    python_exe = get_python_exe(env_path)
    if not python_exe:
        return {"ok": False, "exit_code": 2, "error": "missing python executable"}

    # Discovery of packages to check
    # For now, we scan site-packages for .dist-info
    # A full implementation would use standard library importlib.metadata if available in the target python,
    # but we are running from outside. So manual scan of site-packages is robust for broken envs.
    
    from .discovery import get_site_packages
    site_pkgs = get_site_packages(python_exe)
    if not site_pkgs or not Path(site_pkgs[0]).exists():
         return {"ok": False, "exit_code": 2, "error": "missing site-packages"}
    
    to_check = []
    for sp_path in site_pkgs:
        sp = Path(sp_path)
        if not sp.exists():
            continue
        dists = list(sp.glob("*.dist-info"))
        
        for d in dists:
            # Get import names
            imports = get_toplevel_imports(d)
            for imp in imports:
                if not getattr(args, "full", False):
                    if imp.lower() not in CRITICAL_PACKAGES:
                        continue
                
                # Skip invalid identifiers (e.g. ujson-stubs)
                if not imp.isidentifier():
                    continue
                    
                to_check.append((d.name, imp, sp))
    to_check = sorted(list(set(to_check)))
    
    if not to_check:
         # Fallback: if no critical packages found in lazy mode, warn user or check a few random ones?
         # Or just return empty report.
        if not getattr(args, "full", False):
            # Avoid requiring a translation entry for this experimental path.
            print("No import candidates found in lazy mode.")
        return {"ok": True, "exit_code": 0, "report": {"env": env_path, "checks": 0, "failures": []}}

    results = []
    
    # Parallel execution
    max_workers = min(32, (os.cpu_count() or 1) * 4) 
    print(f"Verifying {len(to_check)} imports using {max_workers} threads...")
    
    progress = Progress(total=len(to_check), label="Verifying imports")
    
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        future_to_check = {
            executor.submit(check_import, import_name, python_exe): (dist_name, import_name, sp_path)
            for dist_name, import_name, sp_path in to_check
        }
        
        completed_count = 0
        for future in concurrent.futures.as_completed(future_to_check):
            dist_name, import_name, sp_path = future_to_check[future]
            try:
                ok, error = future.result()
            except Exception as exc:
                ok, error = False, str(exc)
            
            results.append(
                {
                    "dist": dist_name,
                    "dist_path": sp_path / dist_name,  # store path for fixer
                    "import": import_name,
                    "ok": ok,
                    "error": error,
                }
            )
            
            completed_count += 1
            progress.update(completed_count)

    progress.finish()
    
    failures = [r for r in results if not r["ok"]]

    if not args.json:
        print("\nImport Verification Report:")
        print(f"Checked {len(to_check)} imports in {env_path}\n")
        if not failures:
            print("All checked imports succeeded!")
        else:
            print(f"Found {len(failures)} broken imports:\n")
            for f in failures:
                print(f"  ❌ {f['import']} (from {f['dist']})")
                if f.get("error"):
                    for line in f["error"].splitlines()[:8]:
                        print(f"      {line}")

    fix_report = None
    if getattr(args, "fix", False) and failures:
        fix_report = attempt_fix(
            failures,
            python_exe,
            env_path,
            manager,
            base_prefix=base_prefix,
            debug=show_json_output,
        )

    post_failures = []
    if fix_report is not None:
        plan_items = fix_report.get("plan") or []
        plan_by_dist = {p.get("dist"): p for p in plan_items if isinstance(p, dict) and p.get("dist")}
        for f in failures:
            dist = f.get("dist")
            imp = f.get("import")
            p = plan_by_dist.get(dist) if dist else None
            if p and p.get("kind") in ("skip", "remove"):
                continue
            if not imp or not isinstance(imp, str) or not imp.isidentifier():
                post_failures.append(
                    {"dist": dist, "import": imp, "error": f.get("error") or "invalid import name"}
                )
                continue
            ok, err = check_import(imp, python_exe)
            if not ok:
                post_failures.append({"dist": dist, "import": imp, "error": err})
        if not args.json:
            if post_failures:
                print(f"\nPost-fix: {len(post_failures)} import(s) still failing.")
                for pf in post_failures[:12]:
                    dist = pf.get("dist") or "?"
                    imp = pf.get("import") or "?"
                    print(f"  ❌ {imp} (from {dist})")
            else:
                print("\nPost-fix: all non-skipped imports OK.")
    else:
        post_failures = [
            {"dist": f.get("dist"), "import": f.get("import"), "error": f.get("error")} for f in failures
        ]

    ok_all = len(post_failures) == 0 and (fix_report is None or bool(fix_report.get("ok")))

    report = {
        "env": env_path,
        "python": python_exe,
        "checks": len(to_check),
        "failures": [
            {
                "dist": f.get("dist"),
                "import": f.get("import"),
                "error": f.get("error"),
            }
            for f in failures
        ],
        "post_failures": post_failures,
        "fix": fix_report,
    }
    return {"ok": ok_all, "exit_code": 0 if ok_all else 1, "report": report}



import unittest


class TestAdoptPipUninstallPlan(unittest.TestCase):
    def test_msgpack_alias_uninstalls_when_versions_match(self):
        from env_repair.repair import _adopt_pip_uninstall_plan

        uninstallable, skipped = _adopt_pip_uninstall_plan(
            pip_to_conda={"msgpack": "msgpack-python"},
            pip_versions={"msgpack": "1.1.2"},
            entries=[{"name": "msgpack-python", "version": "1.1.2", "channel": "conda-forge"}],
        )
        self.assertEqual(uninstallable, ["msgpack"])
        self.assertEqual(skipped, [])

    def test_msgpack_alias_keeps_when_versions_differ(self):
        from env_repair.repair import _adopt_pip_uninstall_plan

        uninstallable, skipped = _adopt_pip_uninstall_plan(
            pip_to_conda={"msgpack": "msgpack-python"},
            pip_versions={"msgpack": "1.1.2"},
            entries=[{"name": "msgpack-python", "version": "1.1.1", "channel": "conda-forge"}],
        )
        self.assertEqual(uninstallable, [])
        self.assertEqual(skipped[0]["pip"], "msgpack")


if __name__ == "__main__":
    unittest.main()

